# -*- coding: utf-8 -*-
"""
Created on Fri Dec 18 17:23:00 2020
@author: ju-bar

Binning functions

Note: These are test functions. It turns out that they are too slow
from Python to be of good use. However, they work like a charm and
are fast enough when implemented in C++ language.

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
from numba import jit # include compilation support
from emilys.image.geometry import polygon_area, points_close #, poly_intersect
#
#
def edge_connected(cand1, cand2):
    if (cand1[1] == cand2[1]) and (cand1[3] == cand2[3]): return True
    if (cand1[1] == cand2[2]) and (cand1[3] == cand2[4]): return True
    if (cand1[2] == cand2[1]) and (cand1[4] == cand2[3]): return True
    if (cand1[2] == cand2[2]) and (cand1[4] == cand2[4]): return True
    return False
#
#
class pixelgrid:
    """

    Class pixelgrid stores 2d grid parameters and provides some utility
    geometrical functions.

    Attributes
    ----------
        x0, y0 : float
            grid offset coordinates in a reference frame
        dx, dy : float
            grid spacings in reference frame units
        gamma : float
            angle between grid vectors in radians of the reference frame
        rot : float
            rotation of the first grid vector to the first reference vector
            in radians of the reference frame.
        _matrix : numpy.ndarray, shape(2,2), float
            grid basis matrix
        _inverse : numpy.ndarray, shape(2,2), float
            grid basis inverse matrix

        Note: When changing any attribute manually, better call get_matrix
        before using other utility functions to update _matrix and _inverse

    Members
    -------
        offset : getter and setter for numpy.ndarray, shape (2,), float
            related to the grid offset point (x0, y0)
        matrix : returns the grid matrix
        inverse_matrix: returns the inverse grid matrix
        get_matrix : updates the _matrix values from other parameters
            returns the matrix
        pixel_area : returns the area of one pixel in units
            of the reference frame
        pixel_pos : returns the center position of one pixel identified
            by grid indices
        pixel_fidx : returns fractional grid indices for one position
            given in units of the reference frame
        pixel_corners : returns a list of positions in reference frame
            units that represent the corners of a pixel identified by
            grid indices
        pixel_overlap : returns a list of pixels another grid that
            have intersections with pixels of an input grid. Also returns
            a list with the intersection areas.

    """
    def __init__(self, x0=0., y0=0., dx=1., dy=1., gamma=0.5*np.pi, rot=0.):
        self.x0 = x0
        self.y0 = y0
        self.dx = dx
        self.dy = dy
        self.gamma = gamma
        self.rot = rot
    @property
    def offset(self):
        return np.array([self.x0, self.y0])
    @offset.setter
    def offset(self, value):
        self.x0 = value[0]
        self.y0 = value[1]
    def get_matrix(self):
        mr = np.array([[np.cos(self.rot),-np.sin(self.rot)],[np.sin(self.rot),np.cos(self.rot)]])
        md = np.array([[self.dx,0.],[self.dy * np.cos(self.gamma),self.dy * np.sin(self.gamma)]]).T
        self._matrix = np.round(np.dot(mr,md), 15)
        self._inverse = np.linalg.inv(self._matrix)
        return self._matrix
    @property
    def matrix(self):
        if not hasattr(self,'_matrix'): 
            self.get_matrix()
        return self._matrix
    @property
    def inverse_matrix(self):
        if not hasattr(self,'_inverse'): 
            self._inverse = np.linalg.inv(self.matrix)
        return self._inverse
    @property
    def pixel_area(self):
        return np.linalg.det(self.matrix)
    @property
    def pixel_max_size(self):
        d0 = np.dot(self.matrix, [1.,  1.])
        d1 = np.dot(self.matrix, [1., -1.])
        return max(np.sqrt(d0[0]**2 + d0[1]**2), np.sqrt(d1[0]**2 + d1[1]**2))
    @property
    def pixel_min_size(self):
        mt = self.matrix.T
        ap = np.array([-mt[0,1], mt[0,0]])
        bp = np.array([-mt[1,1], mt[1,0]])
        return min(np.abs(np.dot(bp, mt[0])) / np.sqrt(np.dot(mt[0],mt[0])), np.abs(np.dot(ap, mt[1])) / np.sqrt(np.dot(mt[1],mt[1])))
    def pixel_idx(self, ix, iy, ndimx, ndimy, periodic=False):
        jx = ix; jy = iy
        if periodic: # periodic wrap
            jy = (iy % ndimy); jx = (ix % ndimx)
        else:
            if (ix < 0) or (ix >= ndimx) or (iy < 0) or (iy >= ndimy):
                return -1
        return jx + jy * ndimx
    def idx_pixel(self, idx, ndimx):
        ix = idx % ndimx
        iy = int((idx - ix) / ndimx)
        return [ix, iy]
    def pixel_pos(self, ix, iy):
        return self.offset + np.dot(self.matrix, [ix,iy])
    def pos_pixel(self, pos):
        return np.dot(self._inverse,(pos - self.offset))
    def pixel_fidx(self, pos):
        return np.dot(self.inverse_matrix, (pos - self.offset))
    def pos_in_pixel(self, pos, ix, iy):
        fidx = self.pixel_fidx(pos) - np.array([ix,iy], dtype=float)
        if (fidx[0]>=-0.5 and fidx[0]<0.5 and fidx[1]>=-0.5 and fidx[1]<0.5): return True
        return False
    def npos_in_pixel(self, pos, ix, iy):
        fidx = self.pixel_fidx(pos) - np.array([ix,iy], dtype=float)
        if (fidx[0]>=-0.5 and fidx[0]<0.5 and fidx[1]>=-0.5 and fidx[1]<0.5): return 1
        return 0
    def pixel_corners(self, ix, iy, closed=False):
        l_cor = np.array([[-0.5,-0.5],[0.5,-0.5],[0.5,0.5],[-0.5,0.5]])  + np.array([ix,iy])
        l_ver = np.dot(self.matrix, l_cor.T).T + self.offset
        if closed:
            return np.append(l_ver,l_ver[0]).reshape((len(l_ver)+1,2))
        return l_ver
    def other_pixel_corners_in_pixel(self, pix, other, pix_other):
        m_inside = np.zeros([4,3], dtype=float)
        c_other = other.pixel_corners(pix_other[0], pix_other[1])
        for i in range(0, 4):
            m_inside[i,0] = self.pos_in_pixel(c_other[i],pix[0],pix[1])
            m_inside[i,1:3] = c_other[i]
        return m_inside
    def register_other_edges(self, other, tol=1.E-6):
        m_self = self.matrix.T
        self._v_self = np.array([m_self[0], m_self[1], -m_self[0], -m_self[1]]) # vectors between corners ... this grid
        m_other = other.matrix.T
        self._v_other = np.array([m_other[0], m_other[1], -m_other[0], -m_other[1]]) # vectors between corners ... other grid
        if hasattr(self, '_reg_edge_crossing'): del self._reg_edge_crossing
        self._reg_edge_crossing = np.zeros([4,4], dtype=int)
        if hasattr(self, '_reg_edge_matrix'): del self._reg_edge_matrix
        self._reg_edge_matrix = np.zeros([4,4,2,2], dtype=float)
        for i in range(0,4): # loop over edges of pix in this grid
            for j in range(0,4): # loop over edges of pix_other in other grid
                MDP12 = np.array([self._v_self[i], -self._v_other[j]]).T
                if np.abs(np.linalg.det(MDP12)) > tol: # this means crossing edges
                    self._reg_edge_crossing[i, j] = 1
                    self._reg_edge_matrix[i, j, :, :] = np.linalg.inv(MDP12)
    def other_pixel_to_pixel_edge_crossings(self, pix, other, pix_other, tol=1.E-6):
        if ((not hasattr(self, '_reg_edge_crossing')) or (not hasattr(self, '_reg_edge_matrix'))):
            self.register_other_edges(other, tol)
        c_self = self.pixel_corners(pix[0], pix[1]) # corners of pix in this grid
        c_other = other.pixel_corners(pix_other[0], pix_other[1])
        m_cross = np.zeros([4,4,5], dtype=float) # result matrix of corner x corner x [cross_type, p_x, p_y, t1, t2]
        for i in range(0,4): # loop over edges of pix in this grid
            for j in range(0,4): # loop over edges of pix_other in other grid
                if self._reg_edge_crossing[i, j] > 0: # this means crossing edges
                    DPL = c_other[j] - c_self[i] # distance vector between starting corners
                    t = np.dot(self._reg_edge_matrix[i, j], DPL)
                    if (t[0]>=0.0 and t[0]<=1.0 and t[1]>=0.0 and t[1]<=1.0): # crossing in both edge ranges
                        m_cross[i, j, 0] = 1
                        p = c_self[i] + t[0] * self._v_self[i]
                        m_cross[i, j, 1:3] = p[0:2]
                        m_cross[i, j, 3:5] = t[0:2]
        return m_cross
    def other_pixel_intersect(self, pix, other, pix_other, tol=1.E-6):
        cand = np.zeros([24,5], dtype=int) # candidate vertices index table
        cand_pt = np.zeros([24,2], dtype=float) # candidate positions
        poly = [] # finals intersection polygon
        conn = [0, 0, 0, 0] # flags if edges of the other grid are part of the intersection
        m_cc_1 = self.other_pixel_corners_in_pixel(pix, other, pix_other) # list of other corner pixels with inside this pixel flags
        m_cc_2 = other.other_pixel_corners_in_pixel(pix_other, self, pix) # list of this corner pixels with inside other pixel flags
        m_ce = self.other_pixel_to_pixel_edge_crossings(pix, other, pix_other, tol) # matrix of edge crossing data
        # fill candidate list
        for i in range(0, 4):
            if m_cc_1[i,0] > 0.5: # corners of the other grid pixel that are in this grid pixel -> assigned to grid "1"
                cand[i, :] = [1, 1, 1, i, (i - 1)%4]
                cand_pt[i, 0:2] = m_cc_1[i, 1:3]
            if m_cc_2[i,0] > 0.5: # corners of this grid pixel that are in the other grid pixel -> assigned to grid "0"
                cand[i + 4, :] = [1, 0, 0, i, (i-1)%4]
                cand_pt[i + 4, 0:2] = m_cc_2[i, 1:3]
        for i in range(0, 4):
            for j in range(0, 4):
                if m_ce[i, j, 0] > 0.5: # edge crossing between corners of both pixels
                    idx = 8 + j + i * 4
                    cand[idx,:] = [1, 0, 1, i, j] # on edge index i on this grid pixal and edge index j on the other grid pixel
                    cand_pt[idx, 0:2] = m_ce[i, j, 1:3]
        # run intersection polynomial logics on the candidate list and fill the polynomial
        nc = np.sum(cand[:,0]) # number of candidates
        ic_prev = -1 # index of the previous candidate
        while nc > 0: # still candidates to handle
            for ic in range(0, 24): # loop over candidate array
                ic_add = -1 # new candidate index to add
                if cand[ic,0] > 0: # this is an unhandled candidate
                    if (ic_prev >= 0): # compare candidate ic against the previous for connection
                        if edge_connected(cand[ic], cand[ic_prev]):
                            ic_add = ic
                        # if ((cand[ic,1] == cand[ic_prev,1] and cand[ic,3] == cand[ic_prev,3]) or
                        #     (cand[ic,1] == cand[ic_prev,2] and cand[ic,3] == cand[ic_prev,4]) or
                        #     (cand[ic,2] == cand[ic_prev,1] and cand[ic,4] == cand[ic_prev,3]) or
                        #     (cand[ic,2] == cand[ic_prev,2] and cand[ic,4] == cand[ic_prev,4])):
                        #     ic_add = ic
                    else:
                        ic_add = ic
                if ic_add >= 0: # add the candidate to the intersection polygon
                    poly.append(cand_pt[ic_add])
                    if cand[ic_add,1] == 1: conn[cand[ic_add,3]] = 1
                    if cand[ic_add,2] == 1: conn[cand[ic_add,4]] = 1
                    cand[ic_add, 0] = 0 # no longer a candidate
                    nc -= 1 # decrement candidate count
                    ic_prev = ic_add
                    for jc in range(0, 24): # loop over candidate list and inactivate close points
                        if jc == ic_add: continue # skip current point
                        if 0 == cand[jc, 0]: continue # skip inactive points
                        if points_close(cand_pt[ic_add], cand_pt[jc], tol): # close point -> deactivate
                            cand[jc, 0] = 0
                            nc -= 1 # decrement candidate count
                    break # stop looping ic, new run from while ...
            if ic_add < 0: break # finish as nothing was added
        return (poly, conn)
    def pixel_overlap(self, pix, grid, i_max, r_min=1, periodic=False, debug=False, tol=1.E-6):
        """

        Calculates the overlap of pixel pix of this grid with pixels of the
        input grid assuming that i_max is the maximum index of the input grid.

        Parameters
        ----------
            pix : array of 2 int
                pixel index of this grid
            grid : pixelgrid
                input grid that is compared to this grid
            i_max : array of 2 int
                input grid maximum index (colmns, rows)
            r_min : int, default 1
                minimum search range, values smaller than 0 deactivate search range limit
            periodic : boolean, default: False
                flags use of periodic boundary conditions
            debug : boolean, default: False
                flags debug output via print
            tol : float, default 1.E-6
                tolerated minimum fractional intersection area to consider

        Returns
        -------
            l_pix, l_fac, cover, rng
            l_pix : list of int
                pixel indices when interpreting the input grid as row-wise stream
            l_fac : list of float
                pixel relative intersection factors (normalized to pixel areas)
                these numbers are fractions of input grid pixels covered by the
                pixel pix of this grid
            cover : float
                total relative intersection of pix by the list of pixels l_pix
            rng : int
                index range checked


        """
        fi = grid.pixel_fidx(self.pixel_pos(*pix)) # closest pixel on input grid, fractional index
        ni = np.round(fi,0).astype(int) # closest pixel on input grid, integer index
        nir = 0 # current search radius on grid 1
        nir_min = r_min
        acov = 0.0 # coverage
        dcov = 1.0 # change of coverage
        l_idx = [] # list of grid 1 pixels (result)
        l_fac = [] # list of grid 1 factors (result)
        l_hnd = [] # list of handled pixel indices (internal)
        while (dcov > tol) or (nir < nir_min): # coverage changes
            dcov = 0.
            for i1 in range(ni[1]-nir, ni[1]+nir+1): # loop input grid rows
                if periodic:
                    j1 = i1 % i_max[1] # periodic wrap
                else:
                    if (i1 < 0 or i1 >= i_max[1]):
                        continue # skip out of bounds
                    j1 = i1
                i2 = j1 * i_max[0]
                for i0 in range(ni[0]-nir, ni[0]+nir+1): # loop input grid columns
                    if periodic:
                        j0 = i0 % i_max[0] # periodic wrap
                    else:
                        if (i0 < 0 or i0 >= i_max[0]):
                            continue # skip out of bounds
                        j0 = i0
                    idx1 = j0 + i2 # grid pixel index
                    if idx1 < 0: continue
                    if idx1 in l_hnd: continue # already handled, skip
                    l_hnd.append(idx1) # mark this pixel as handled
                    l_vf, con = self.other_pixel_intersect(pix, grid, [j0, j1], tol=tol)
                    # l_vf = poly_intersect(grid.pixel_corners(i0,i1), self.pixel_corners(*pix), tol=tol, debug=debug) # get intersection
                    if len(l_vf) > 0: # there is intersection
                        nir_min = -1 # deactivate external search range pushing
                        afac = polygon_area(l_vf) / grid.pixel_area
                        if (afac > tol):
                            acov += afac
                            dcov += afac
                            l_idx.append(idx1)
                            l_fac.append(afac)
                            if debug: print('dbg (pixel_overlap): added pixel [{:d},{:d}] ({:d}) with overlap: {:.6G}'.format(i0, i1, idx1, afac))
            nir +=1
        return (l_idx, l_fac, acov, nir)
#
#
@jit
def hfac_rebin(a, a_hash_idx, a_hash_fac):
    """

    Rebins array a to a new array by a given factor hash lists.
    The length of the output array is given by the length of the hash lists.
    Each item of a_hash_idx is a list of indices identifying items of a
    that contribute to an item of the output. Each item of a_hash_fac is
    a list of factors determining by how much an item of a contributes.

    Parameters
    ----------

        a : numpy.ndarray of shape (n)
            input array
        a_hash_idx : list
            list of lists of indices
        a_hash_fac : list
            list of lists of factors

    Returns
    -------
        numpy.ndarray of shape (len(a_hash_idx))

    Remarks
    -------
        This is meant for flattened input and output arrays.

    """
    n = len(a)
    m = len(a_hash_idx)
    b = np.zeros(m, dtype=a.dtype)
    for j in range(0, m):
        for i in range(0, len(a_hash_idx[j])):
            idx = a_hash_idx[j][i]
            if (idx >= 0 and idx < n): b[j] += (a_hash_fac[j][i] * a[idx])
    return b


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 11 10:04:00 2021
@author: ju-bar

Correlation utilities

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np

def corr_norm(a, b):
    """
    Calculates the normalized correlation coefficient of two arrays.

    Parameters
    ----------
        a : numpy array
            input array
        b : numpy array
            input array to be correlated with a

    Returns
    -------
        float
            correlation coefficient [-1., 1.]
    """
    fa = a.flatten().astype(np.dtype('f8'))
    fb = b.flatten().astype(np.dtype('f8'))
    assert len(fa) == len(fb), 'This is for equal size arrays only.'
    ma = np.mean(fa)
    fa_0 = fa - ma
    ssa = np.sum(fa_0**2)
    mb = np.mean(fb)
    fb_0 = fb - mb
    ssb = np.sum(fb_0**2)
    if ssb * ssa == 0.: return 0.
    return np.dot(fa_0/np.sqrt(ssa), fb_0/np.sqrt(ssb))


# -*- coding: utf-8 -*-
"""
Created on Mon Jun 20 14:44:00 2022
@author: ju-bar

Functions handling input and output of structure data
via the CIF file format.

The CEL file format is a text format used here to define a volume
in space which is filled by atoms.

CIF file structure is interpreted according to the CIF dictionary
version 2.4: ftp://ftp.iucr.org/pub/cif_core.dic
See also: https://www.iucr.org/resources/cif/cif2
The main purpose is for I/O of crystal structures, additional
parameter handling will be implemented later if needed.

Example:
# CIF file example for the structure of perovskite SrTiO3 (from ICSD #80874)
_cell_length_a 3.901
_cell_length_b 3.901
_cell_length_c 3.901
_cell_angle_alpha 90.
_cell_angle_beta 90.
_cell_angle_gamma 90.
_symmetry_space_group_name_H-M 'P m -3 m'
_symmetry_Int_Tables_number 221
loop_
_symmetry_equiv_pos_site_id
_symmetry_equiv_pos_as_xyz
1 'z, y, -x'
2 'y, x, -z'
3 'x, z, -y'
4 'z, x, -y'
5 'y, z, -x'
6 'x, y, -z'
7 'z, -y, x'
8 'y, -x, z'
9 'x, -z, y'
10 'z, -x, y'
11 'y, -z, x'
12 'x, -y, z'
13 '-z, y, x'
14 '-y, x, z'
15 '-x, z, y'
16 '-z, x, y'
17 '-y, z, x'
18 '-x, y, z'
19 '-z, -y, -x'
20 '-y, -x, -z'
21 '-x, -z, -y'
22 '-z, -x, -y'
23 '-y, -z, -x'
24 '-x, -y, -z'
25 '-z, -y, x'
26 '-y, -x, z'
27 '-x, -z, y'
28 '-z, -x, y'
29 '-y, -z, x'
30 '-x, -y, z'
31 '-z, y, -x'
32 '-y, x, -z'
33 '-x, z, -y'
34 '-z, x, -y'
35 '-y, z, -x'
36 '-x, y, -z'
37 'z, -y, -x'
38 'y, -x, -z'
39 'x, -z, -y'
40 'z, -x, -y'
41 'y, -z, -x'
42 'x, -y, -z'
43 'z, y, x'
44 'y, x, z'
45 'x, z, y'
46 'z, x, y'
47 'y, z, x'
48 'x, y, z'
loop_
_atom_type_symbol
_atom_type_oxidation_number
Sr2+ 2
Ti4+ 4
O2- -2
loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_symmetry_multiplicity
_atom_site_Wyckoff_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
_atom_site_B_iso_or_equiv
_atom_site_occupancy
_atom_site_attached_hydrogens
Sr1 Sr2+ 1 a 0 0 0 . 1. 0
Ti1 Ti4+ 1 b 0.5 0.5 0.5 . 1. 0
O1 O2- 3 c 0 0.5 0.5 . 1. 0
loop_
_atom_site_aniso_label
_atom_site_aniso_type_symbol
_atom_site_aniso_U_11
_atom_site_aniso_U_22
_atom_site_aniso_U_33
_atom_site_aniso_U_12
_atom_site_aniso_U_13
_atom_site_aniso_U_23
Sr1 Sr2+ 0.00789(3) 0.00789(3) 0.00789(3) 0 0 0
Ti1 Ti4+ 0.00558(9) 0.00558(9) 0.00558(9) 0 0 0
O1 O2- 0.00472(38) 0.0111(3) 0.0111(3) 0 0 0
#End of data

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import re
import numpy as np
from emilys.structure.supercell import supercell
import emilys.structure.atom as ato
import emilys.structure.atomtype as aty

d_cif_types = { # CIF data block type dictionary
    '_cell_length_a' : float(1.0),
    '_cell_length_b' : float(1.0),
    '_cell_length_c' : float(1.0),
    '_cell_angle_alpha' : float(1.0),
    '_cell_angle_beta' : float(1.0),
    '_cell_angle_gamma' : float(1.0),
    '_symmetry_space_group_name_H-M' : str('P 1'),
    '_symmetry_space_group_name_H-M_alt' : str('P 1'),
    '_space_group_name_H-M_alt' : str('P 1'),
    '_symmetry_Int_Tables_number' : int(1),
    '_space_group_IT_number' : int(1),
    '_symmetry_equiv_pos_site_id' : int(1),
    '_symmetry_equiv_pos_as_xyz' : str('x,y,z'),
    '_space_group_symop_id' : int(1),
    '_space_group_symop_operation_xyz' : str('x,y,z'),
    '_atom_type_symbol' : str('Fe'),
    '_atom_type_oxidation_number' : float(1.0),
    '_atom_site_label' : str('Fe'),
    '_atom_site_type_symbol' : str('Fe'),
    '_atom_site_type_charge' : float(1.0),
    '_atom_site_fract_x' : float(1.0),
    '_atom_site_fract_y' : float(1.0),
    '_atom_site_fract_z' : float(1.0),
    '_atom_site_B_iso_or_equiv' : float(1.0),
    '_atom_site_U_iso_or_equiv' : float(1.0),
    '_atom_site_occupancy' : float(1.0),
    '_atom_site_adp_type' : str('Uiso'),
    '_atom_site_thermal_displace_type' : str('Uiso'),
    '_atom_site_aniso_type_symbol' : str('Fe'),
    '_atom_site_aniso_U_11' : float(1.0),
    '_atom_site_aniso_U_22' : float(1.0),
    '_atom_site_aniso_U_33' : float(1.0),
    '_atom_site_aniso_U_12' : float(1.0),
    '_atom_site_aniso_U_13' : float(1.0),
    '_atom_site_aniso_U_23' : float(1.0),
    '_atom_site_aniso_B_11' : float(1.0),
    '_atom_site_aniso_B_22' : float(1.0),
    '_atom_site_aniso_B_33' : float(1.0),
    '_atom_site_aniso_B_12' : float(1.0),
    '_atom_site_aniso_B_13' : float(1.0),
    '_atom_site_aniso_B_23' : float(1.0)
}

d_cif_tables = {
    'symmetry_equiv_pos' : {
        'threshold' : 1.0,
        '_symmetry_equiv_pos_site_id' : 0.4,
        '_symmetry_equiv_pos_as_xyz' : 1.0,
        '_space_group_symop_id' : 0.4,
        '_space_group_symop_operation_xyz' : 1.0,
    },
    'atom_type' : {
        'threshold' : 1.0,
        '_atom_type_symbol' : 1.0,
        '_atom_type_oxidation_number' : 0.5
    },
    'atom_site' : {
        'threshold' : 4.0,
        '_atom_site_label' : 1.0,
        '_atom_site_type_symbol' : 1.0,
        '_atom_site_symmetry_multiplicity' : 0.1,
        '_atom_site_Wyckoff_symbol' : 0.1,
        '_atom_site_fract_x' : 1.0,
        '_atom_site_fract_y' : 1.0,
        '_atom_site_fract_z' : 1.0,
        '_atom_site_B_iso_or_equiv' : 1.0,
        '_atom_site_U_iso_or_equiv' : 1.0,
        '_atom_site_adp_type' : 0.1,
        '_atom_site_thermal_displace_type' : 0.1,
        '_atom_site_occupancy' : 1.0,
        '_atom_site_attached_hydrogens' : 0.1
    },
    'atom_site_aniso' : {
        'threshold' : 7.0,
        '_atom_site_aniso_label' : 1.0,
        '_atom_site_aniso_type_symbol' : 1.0,
        '_atom_site_aniso_U_11' : 1.0,
        '_atom_site_aniso_U_22' : 1.0,
        '_atom_site_aniso_U_33' : 1.0,
        '_atom_site_aniso_U_12' : 1.0,
        '_atom_site_aniso_U_13' : 1.0,
        '_atom_site_aniso_U_23' : 1.0
    }
}

str_CIF_char_qm = "'\"" # quotation marks
str_CIF_char_sep = " " # separators
str_CIF_char_com = "#" # comments
str_CIF_symop_ch = "xyz" # symmetry operation order

def get_CIF_float(s):
    val = 0.0
    rs = s.strip()
    if len(rs) > 0:
        if rs != '.' and rs != '?':
            ib = rs.find('(')
            if ib > 0: rs = rs[0:ib]
            val = float(rs)
    return val

def get_CIF_int(s):
    val = int(0)
    rs = s.strip()
    if len(rs) > 0:
        if rs != '.' and rs != '?':
            ib = rs.find('(')
            if ib > 0: rs = rs[0:ib]
            val = int(rs)
    return val

def sep_CIF_str(s):
    """

    Separates the first sub-string in s from the rest.
    This takes quotation marks into account.

    Parameters
    ----------

        s : str
            input string

    Results
    -------

        str, str
            The first substring and the remaining part.

    """
    rs = s.strip() # remove leading and trailing spaces and make copy of the input
    iqm1 = str_CIF_char_qm.find(rs[0]) # check if the first sub-string begins by one of the quotation marks
    if iqm1 >= 0: # it does ...
        iqm2 = rs[1:].find(str_CIF_char_qm[iqm1]) # .. find the closing mark
        if iqm2 >= 0: # .. found
            s1 = rs[1:(iqm2+1)] # copy the substring enclosed by quotation marks (marks removed)
            s2 = rs[(iqm2+2):].strip() # get the rest and strip
            return s1, s2 # return the two parts
    isep = rs.find(str_CIF_char_sep) # find the next separator (only one separator)
    if isep >= 0: # .. found
        s1 = rs[0:isep] # take the part in front of the separator as first sub-string
        s2 = rs[isep:].strip() # take the rest and strip
    else: # no separator, assume only one sub-string
        s1 = rs # 
        s2 = "" # empty rest
    return s1, s2

def read_CIF_scalar(s):
    """

    Reads a CIF pair of parameter name and value from the string s

    Parameters
    ----------

        s : str
            input string with a set of substrings where the second might
            be inside quotation marks

    Returns
    -------

        list : dtype = str, len = 2
            The pair of parameter name and value in form of str

    """
    l_dat = []
    rs = s.strip()
    while len(rs) > 0:
        s1, s2 = sep_CIF_str(rs)
        l_dat.append(s1)
        rs = s2
    while len(l_dat) < 2: # expand to contain n items
        l_dat.append('')
    return l_dat[0:2] # return only a list of length 2

def read_CIF_table_row(s, n):
    """

    Reads at maximum n values from the input string s

    Parameters
    ----------
        
        s : str
            input string with a set of substrings that might
            be inside quotation marks
        n : int
            number of substrings expected

    Returns
    -------

        list : dtype = str
            list of strings representing values of one table row

    """
    l_dat = []
    rs = s.strip()
    while len(rs) > 0:
        s1, s2 = sep_CIF_str(rs)
        l_dat.append(s1)
        rs = s2
    while len(l_dat) < n: # expand to contain n items
        l_dat.append('')
    return l_dat

def parse_CIF_symop_cmp(s):
    """

    Determines a vector and a scalar from the input string defining
    one line of symmetry operation in terms of a linear transformation.

    Parameters
    ----------

        s : str
            Input string, e.g. 'x' or '-x+1/2' or 'x-y'

    Returns
    -------

        np.array of shape=(4)
            The first 3 numbers are a 3-tuple defining a row of a matrix
            and the fourth number is the component of a shift vector

    """
    v = np.zeros(4, dtype=float)
    rs = s.strip().replace(' ','') # remove any whitespace character
    l = len(rs)
    if l > 2: # check for an operation -> bifurcate (recurse)
        i_op = 1 + rs[1:].find('+') # addition of two sub-terms
        if i_op > 0:
            v1 = parse_CIF_symop_cmp(rs[0:i_op])
            v2 = parse_CIF_symop_cmp(rs[i_op+1:l])
            return v1 + v2
        i_op = 1 + rs[1:].find('-') # subtraction of two sub-terms
        if i_op > 0:
            v1 = parse_CIF_symop_cmp(rs[0:i_op])
            v2 = parse_CIF_symop_cmp(rs[i_op+1:l])
            return v1 - v2
        i_op = 1 + rs[1:].find('*') # multiplication of two sub-terms
        if i_op > 0:
            v1 = parse_CIF_symop_cmp(rs[0:i_op])
            v2 = parse_CIF_symop_cmp(rs[i_op+1:l])
            # this handles cases like 2*x or 3*6
            # no second order forms, no brackets
            v[0:3] = v1[0:3] * v2[3] + v2[0:3] * v1[3]
            v[3] = v1[3] * v2[3]
            return v
        i_op = 1 + rs[1:].find('/') # multiplication of two sub-terms
        if i_op > 0:
            v1 = parse_CIF_symop_cmp(rs[0:i_op])
            v2 = parse_CIF_symop_cmp(rs[i_op+1:l])
            # this handles cases like x/2 or 1/3
            # no reciprocals, only division by numbers
            v[0:3] = v1[0:3] / v2[3]
            v[3] = v1[3] / v2[3]
            return v
    else: # trivial case, solve here (end of recurse level)
        i_ch = -1
        j = -1
        for i in range(0,3):
            j = rs.find(str_CIF_symop_ch[i])
            if j >= 0: # this channel is present
                i_ch = i # remember which character
                rs = rs[:j] + '1' + rs[(j+1):] # replace character by number
                break # stop looping
        # rs is just '1' or '-1' now
        vt = float(rs)
        if i_ch == -1:
            v[3] = vt
        else:
            v[i_ch] = vt
    return v

def get_CIF_symop(s):
    """

    Transforms a string to a linear operation consisting of
    a 3x3 matrix and a 3D-vector.

    Parameters
    ----------

        s : str
            Operation text, e.g. 'x,y,z' or '-x, y, -z + 1/2'

    Returns
    -------

        numpy.ndarray : dtype=float, shape=(4,3)
            3x3 matrix in the first 3 rows, 3D vector in the fourth row

    """
    v = np.zeros((4,3), dtype=float)
    rs = s.strip()
    l = len(rs)
    c1 = rs.find(',')
    c2 = rs.rfind(',')
    if ((c1 > 0) and (c2 < l-1) and (c2-c1 > 1)): # valid string with 2 commas
        s1 = rs[:c1]
        s2 = rs[(c1+1):(c2)]
        s3 = rs[c2+1:]
        v1 = parse_CIF_symop_cmp(s1)
        v2 = parse_CIF_symop_cmp(s2)
        v3 = parse_CIF_symop_cmp(s3)
        v[0,0:3] = v1[0:3]
        v[1,0:3] = v2[0:3]
        v[2,0:3] = v3[0:3]
        v[3,0] = v1[3]
        v[3,1] = v2[3]
        v[3,2] = v3[3]
    return v

def read_CIF(file, debug=False):
    """

    Reads a file and interprets content as atomic structure data
    given in the CIF format and returns a dictionary of the data.

    Parameters
    ----------

        file : str
            File name
        debug : boolean
            Switch debug output

    Returns
    -------

        dict
            Structure data dictionary.

    """
    d_cif = {}
    assert type(file) is str, "This expects a string as input parameter."
    # open the file and read the lines
    with open(file) as file_in:
        if debug: print('dbg (read_cif): opened file [' + file + ']')
        lines = []
        for line in file_in:
            icomm = line.find('#')
            if icomm >= 0:
                lines.append(line[0:icomm].strip()) # add lines without comment and stripped
            else:
                lines.append(line.strip()) # add lines without leading and trailing spaces
        file_in.close()
    if debug: print('dbg (read_cif): read', len(lines),'lines of text.')
    # setup book-keeper for line handling
    l_handled = np.zeros(len(lines), dtype=int)
    # find all tables "loop_"
    d_tables = {}
    for il in range(0, len(lines)): # loop over lines numerically by index
        if 0 == len(lines[il]): 
            l_handled[il] = 1
            continue # skip empty lines
        if 0 == lines[il].find('loop_'): # handle loop
            num_tab = len(d_tables.keys())
            str_tab = str(num_tab)
            d_tables[str_tab] = {
                "line_start" : il,
                "column" : [],
                "data" : []
            } # insert new table named by number
            l_handled[il] = 1
    if debug: print('dbg (read_cif): found', len(d_tables.keys()), 'tables')
    # get columns of all tables and read the table values as strings
    for str_tab in d_tables.keys():
        d_table = d_tables[str_tab]
        il = d_table['line_start'] + 1 # expect column names to start after "loop_"
        while 0 == lines[il].find('_'): # get column keys
            d_table['column'].append(lines[il]) # add column name to list
            il += 1 # next line
        num_col = len(d_table['column'])
        b_data_line = True
        while b_data_line:
            if il >= len(lines): break
            if (0 == lines[il].find('loop_')) or (0 == lines[il].find('_')) or (0 == len(lines[il])): b_data_line = False
            if b_data_line:
                d_table['data'].append(read_CIF_table_row(lines[il], num_col))
            il += 1 # next line
        num_row = len(d_table['data'])
        d_table['line_end'] = d_table['line_start'] + num_col + num_row
        l_handled[d_table['line_start']:d_table['line_end']+1] = 1
        if debug: print('dbg (read_cif): table #' + str_tab + ': #columns:', num_col, '#rows:', num_row)
    # identify table content and name it
    for str_tab in d_tables.keys():
        d_table = d_tables[str_tab]
        d_table['name'] = 'unknown' # preset unknown table data
        l_str_cif_tables = list(d_cif_tables.keys())
        l_vote = np.zeros(len(l_str_cif_tables), dtype=float) # votes from cif_tables
        for icol in range(0, len(d_table['column'])): # loop over columns of the loaded table
            str_col = d_table['column'][icol] # get column name
            for i_cif_table in range(0, len(l_str_cif_tables)): # loop over known CIF table names
                str_cif_table = l_str_cif_tables[i_cif_table]
                if str_col in d_cif_tables[str_cif_table].keys(): # check if the currently checked CIF table contains the column name
                    l_vote[i_cif_table] += d_cif_tables[str_cif_table][str_col] # ... yes, increment vote by name weight
        for i_cif_table in range(0, len(l_str_cif_tables)): # loop over votes
            str_cif_table = l_str_cif_tables[i_cif_table] # get name
            if l_vote[i_cif_table] >= d_cif_tables[str_cif_table]['threshold']: # vote exceeds threshold
                d_table['name'] = str_cif_table # set table name
                if debug: print('dbg (read_cif): table #' + str_tab + ' identified as ' + str_cif_table)
                break # do not look for a second table
    # add tables to the cif dictionary
    d_cif['tables'] = d_tables
    if debug: print('dbg (read_cif):', sum(l_handled), 'of', len(lines), 'lines handled after table reading.')
    # read scalar parameters from unhandled lines
    for il in range(0, len(lines)):
        if l_handled[il] == 0: # unhandled line
            str_prm_name, str_prm_val = read_CIF_scalar(lines[il])
            if str_prm_name in d_cif_types:
                d_cif[str_prm_name] = str_prm_val
            l_handled[il] = 1
    if debug: print('dbg (read_cif):', sum(l_handled), 'of', len(lines), 'lines handled after scalar parameter reading.')
    return d_cif

def write_CIF_qm(s):
    """
    """
    rs = str(s)
    if rs.find(' ') >= 0: # contains spaces?
        if rs.find("'") >= 0: # contains single qm?
            return '"' + rs + '"' # put in souble quotes
        else:
            return "'" + rs + "'" # put in single quotes
    return rs

def write_CIF(d_cif, file, debug=False):
    """

    Writes a CIF file from dictionary information.

    Parameters
    ----------

        d_cif : dict
            Dictionary with CIF information
        file : str
            File name
        debug : boolean, default=False
            Switch debug print

    Returns
    -------

        int
            Error code.

    """
    io_err = 0
    assert isinstance(file, str), 'This expects that file is input of type str'
    with open(file, "w") as file_out:
        file_out.write("# (cifio)\n")
        for str_CIF_key in d_cif:
            if str_CIF_key == 'tables': continue # skip tables
            file_out.write(str_CIF_key + " " + write_CIF_qm(d_cif[str_CIF_key]) + "\n")
        file_out.write("\n")
        if 'tables' in d_cif:
            for str_CIF_tab in d_cif['tables']:
                d_table = d_cif['tables'][str_CIF_tab]
                if d_table['name'] == 'unknown':
                    file_out.write("# table\n")
                else:
                    file_out.write("# table " + str(d_table['name']) + "\n")
                file_out.write("loop_\n")
                for str_col in d_table['column']:
                    file_out.write(str_col + "\n")
                for l_row in d_table['data']:
                    ll = list(l_row)
                    for i in range(0, len(ll)):
                        ll[i] = write_CIF_qm(ll[i])
                    file_out.write(' '.join(ll) + "\n")
                file_out.write("\n")
        file_out.write("#End_(cifio)\n")
    return io_err

def get_CIF_atom_sites(d_cif):
    l_as = []
    if 'tables' in d_cif:
        d_as = {}
        d_at = {}
        d_aniso = {}
        for str_tab in d_cif['tables']:
            if d_cif['tables'][str_tab]['name'] == 'atom_site':
                d_as = d_cif['tables'][str_tab]
            if d_cif['tables'][str_tab]['name'] == 'atom_type':
                d_at = d_cif['tables'][str_tab]
            if d_cif['tables'][str_tab]['name'] == 'atom_site_aniso':
                d_aniso = d_cif['tables'][str_tab]
        if ('data' in d_as) and ('column' in d_as):
            l_col = list(d_as['column'])
            for asite in d_as['data']:
                a = ato.atom()
                str_key = '_atom_site_type_symbol'
                i_key = l_col.index(str_key)
                if i_key < 0:
                    print('Error (get_CIF_atom_sites): Missing key [' + str_key + '] in table of atom sites.')
                    break
                str_symb = str(asite[i_key])
                a.Z = aty.Z_from_symb(str_symb)
                str_key = '_atom_site_type_occupancy'
                i_key = l_col.index(str_key)
                if i_key >= 0:
                    a.occ = get_CIF_float(asite[i_key])
                else:
                    a.occ = float(1.0)
                # read usio
                a.uiso = float(0.) # initialize with uiso not set
                # - check for specific adp type definition
                str_adp_type = "unknown"
                str_key = '_atom_site_adp_type'
                i_key = l_col.index(str_key)
                if i_key >= 0:
                    str_adp_type = str(asite[i_key])
                str_key = '_atom_site_thermal_displace_type'
                i_key = l_col.index(str_key)
                if i_key >= 0:
                    str_adp_type = str(asite[i_key])
                # - check for Biso column
                str_key = '_atom_site_B_iso_or_equiv'
                i_key = l_col.index(str_key)
                if i_key >= 0:
                    val = get_CIF_float(asite[i_key])
                    if str_adp_type == "unknown":
                        str_adp_type = "Biso"
                    if str_adp_type == "Uiso":
                        a.uiso = val
                    elif str_adp_type == "Biso":
                        a.uiso = val / (8. * np.pi**2) # translate from Biso input to Uiso stored
                # - check for Uiso column
                str_key = '_atom_site_U_iso_or_equiv'
                i_key = l_col.index(str_key)
                if i_key >= 0:
                    if str_adp_type == "unknown":
                        str_adp_type = "Uiso"
                    val = get_CIF_float(asite[i_key]) # uiso input
                    if str_adp_type == "Uiso":
                        a.uiso = val
                    elif str_adp_type == "Biso":
                        a.uiso = val / (8. * np.pi**2) # translate from Biso input to Uiso stored
                # - check anisotropic table and override if available for current atom type
                
                # read charge (oxidation state if possible)
                if ('data' in d_at) and ('column' in d_at):
                    l_col_type = list(d_at['column'])
                    for atype in d_at['data']:
                        str_key_type = '_atom_type_symbol'
                        i_key_type = l_col_type.index(str_key_type)
                        str_symb_type = str(atype[i_key_type])
                        if str_symb_type == str_symb: # found type match?
                            str_key_type = '_atom_type_oxidation_number'
                            i_key_type = l_col_type.index(str_key_type)
                            if i_key_type >= 0:
                                a.charge = get_CIF_float(atype[i_key_type])
                            else:
                                a.charge = float(0.)
                            break
                else:
                    a.charge = float(0.)
        else: # problem, no sites
            print('Error (get_CIF_atom_sites): No table of atom sites.')
    return l_as

def get_CIF_atom_sites_P1(d_cif):
    l_as_P1 = []
    l_as = get_CIF_atom_sites(d_cif) # get list of atom sites using atom objects
    return l_as_P1

def CIF_to_supercell(d_cif):
    sc = supercell()
    sc.a0[0] = float(d_cif['_cell_length_a'])
    sc.a0[1] = float(d_cif['_cell_length_b'])
    sc.a0[2] = float(d_cif['_cell_length_c'])
    sc.angles[0] = float(d_cif['_cell_angle_alpha'])
    sc.angles[1] = float(d_cif['_cell_angle_beta'])
    sc.angles[2] = float(d_cif['_cell_angle_gamma'])
    sc.basis = sc.get_basis()
    sc.l_atoms = get_CIF_atom_sites_P1(d_cif) # get the atom site list in space group P1
    return sc


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 30 09:44:00 2021
@author: ju-bar

Modified on Mon Jun 20 11:50:00 2022 (ju-bar) adding additional data dict

Functions and classes handling supercells of atomic structures

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

import os, sys
import numpy as np
import emilys.structure.atomtype as aty
from emilys.structure.atom import atom, get_str_from_charge
from copy import deepcopy

class supercell:
    """

    class supercell

    Parameters
    ----------

        a0 : numpy.ndarray([a,b,c], dtype=float)
            cell lattice constants in Angst
        angles : numpy.ndarray([alpha, beta, gamma], dtype=float)
            cell lattice angles between [bc,ca,ab]
        basis : numpy.ndarray([[ax,ay,az],[bx,by,bz],[cx,cy,cz]], dtype=float)
            call basis vectors in Angst
        l_atoms : list of atom objects
            atoms contained in the super cell
        d_add : dictionary
            additional data, depends on I/O routines

    Methods
    -------

        copy():
            Returns a copy of the supercell object.

        report(num_atoms_max):
            Prints a short report of the supercell parameters.

        get_basis():
            Calculates basis vectors from parameters a0 and angles.
            Returns a 3 x 3 numpy array where rows are basis vectors.

        get_composition_str():
            Returns a string representing the atom content of the supercell.

        grow(add_size, center):
            Grows the supercell by given amount in Angst in three dimensions.
            This effectively adds empty space at the large coordinate ends
            of the box. The center option can be used to center the previous
            content in the new box.

        get_avg_pos(l_atoms_idx, proximity, periodic):
            Returns the average position of a list of atoms given by the
            index list l_atoms_idx. Only atoms closer than the proximity 
            parameter will be included and this may be checked under
            periodic boundary conditions.

        get_type_dict(l_atoms_idx):
            Returns a dictionary listing atomic types and sites assigned
            to these types for all atoms indexed in list l_atoms_idx.

        add_atom(Z, uiso, pos, occ, charge):
            Adds an atom to the structure with given parameters.

        keep_atoms(l_atoms_idx):
            Removes all atoms which are not indexed in list l_atoms_idx.

        delete_atoms(l_atoms_idx):
            Removes all atoms which are indexed in list l_atoms_idx.

        periodic():
            Applies periodic boundary conditions to all atoms such that
            their fractional coordinates are >=0 and <1.

        set_uiso(l_atoms_idx, uiso):
            Sets the uiso parameter of all atoms indexed by list
            l_atoms_idx to the given value.

        set_biso(l_atoms_idx, biso):
            Sets the uiso parameter of all atoms indexed by list
            l_atoms_idx by translating the given biso value.
            uiso = biso / (8 Pi**2)

        set_occ(l_atoms_idx, occ):
            Sets the occupancy parameter of all atoms indexed by list
            l_atoms_idx to the given value.

        shift_atoms(l_atoms_idx, shift, periodic):
            Shifts atoms indexed in list l_atoms_idx by a shift vector
            in fractional coordinates.

        shift_all_atoms(shift, periodic):
            Shifts all atoms by a shift vector in fractional coordinates.

        shift_atoms_to(l_atoms_idx, pos, fraction, confinement, mode):
            Shifts atoms indexed in list l_atoms_idx towards pos by
            a given fraction of the initial distance. Shifts can be confined
            to be parallel to planes or lines.

        list_positions(l_atoms_idx):
            Returns a list of positions of atoms identified by index.

        list_close_atoms(l_atoms_idx, proximity, periodic):
            Returns a list of lists of atoms, which are closer than the
            proximity parameter in nanometers. The periodic option switches
            the check of proximity under periodic boundary conditions.

        remove_close_atoms(l_atoms_idx, proximity):
            Returns a list of atom indices to be removed from l_atoms_idx.
            The removel is not performed, so that l_atoms_idx remains 
            unchanged by this routine.

        list_atoms_in_range(dic_range):
            Returns a list of atoms which parameters fall into all range
            specifications listed in the dictionary dic_range. See the
            function definition on how to setup the dictionary.

    """

    def __init__(self):
        self.a0 = np.array([1., 1., 1.]) # lattice constants [x, y, z]
        self.angles = np.array([90., 90., 90.]) # cell angles [alpha, beta, gamma] between [yz, zx, xy]
        self.basis = np.array([[1.,0.,0.],[0.,1.,0.],[0.,0.,1.]]) # list of basis vectors x, y, z
        self.l_atoms = [] # list of atoms
        self.d_add = {} # additional data

    def copy(self):
        return deepcopy(self)

    def report(self, num_atoms_max=10):
        """

        Prints a report of the supercell parameters.

        Parameters
        ----------

            num_atoms_max : int, default 10
                Maximum number of contained atoms to be printed.

        """
        n = len(self.l_atoms)
        print('lattice constants [A]: a = {:.5f}, b = {:.5f}, c = {:.5f}'.format(self.a0[0],self.a0[1],self.a0[2]))
        print('lattice angles [deg]: alpha = {:.4f}, beta = {:.4f}, gamma = {:.4f}'.format(self.angles[0],self.angles[1],self.angles[2]))
        print('number of atoms: {:d}'.format(n))
        n_max = min(n, num_atoms_max)
        if n_max:
            for i in range(0, n_max):
                ato = self.l_atoms[i]
                symb = aty.atom_type_symbol[ato.Z] + get_str_from_charge(ato.charge)
                print('#{:d}: '.format(i) + symb +
                    ', pos = [{:.5f}, {:.5f}, {:.5f}]'.format(ato.pos[0], ato.pos[1], ato.pos[2]) +
                    ', occ = {:.5f}, uiso = {:.5f}'.format(ato.occ, ato.uiso) )

    def get_basis(self):
        """

        Returns the list of three 3D basis vectors defining the lattice.

        """
        l_ac = np.cos( np.deg2rad(self.angles) ).round(15)
        l_as = np.sin( np.deg2rad(self.angles) ).round(15)
        assert np.abs(l_as[2]) > 0., 'cell angle gamma is inavlid'
        return np.array([
                [self.a0[0], self.a0[1] * l_ac[2], self.a0[2] * l_ac[1]],
                [0. , self.a0[1] * l_as[2], self.a0[2] * ( l_ac[0] - l_ac[1] * l_ac[2] ) / l_as[2] ],
                [0. , 0. , self.a0[2] * np.sqrt( l_as[2]**2 - l_ac[1]**2 - l_ac[0]**2 + 2. * l_ac[0] * l_ac[1] * l_ac[2] ) / l_as[2] ]
            ])
    
    def get_composition_str(self):
        """

        Returns a string reflecting the composition of the supercell.

        """
        s_cmp = ''
        d_cmp = {}
        for ato in self.l_atoms:
            symb = aty.atom_type_symbol[ato.Z] + get_str_from_charge(ato.charge)
            if symb in d_cmp:
                d_cmp[symb] += ato.occ
            else:
                d_cmp[symb] = ato.occ
        n_cmp = 0
        for comp in d_cmp:
            if n_cmp > 0: s_cmp += ' ' 
            s_cmp += comp + '_'
            v_occ = d_cmp[comp]
            if v_occ > int(v_occ):
                s_cmp += '{:.2f}'.format(v_occ)
            else:
                s_cmp += str(int(v_occ))
            n_cmp += 1
        return s_cmp

    def grow(self, add_size, center=False):
        """

        Grows the supercell by given amount in Angst in three dimensions.
        This effectively adds empty space at the large coordinate ends
        of the box. The center option can be used to center the previous
        content in the new box.

        Parameters
        ----------

            add_size : numpy.ndarray([x, y, z], dtype=float)
                Size added to the current box dimensions in Angst for
                each lattice dimension.

            center : boolean (default: False)
                If True, the previous content will be centered in the new
                bos. If False, the box will be extended towards the positive
                lattice axes.

        Returns
        -------

            numpy.ndarray([x, y, z], dtype=float) : new size of the box

        """
        new_a0 = self.a0 + add_size # new box size
        f_scale = self.a0 / new_a0 # scaling factor from old to new box size keeping actual positions
        n_atoms = len(self.l_atoms)
        if n_atoms > 0:
            for i in range(0, n_atoms):
                self.l_atoms[i].pos = self.l_atoms[i].pos * f_scale # scale fractional atom positions
        self.a0 = new_a0 # update bos size
        if center: # center previous content in new box?
            f_shift = 0.5 * add_size / self.a0 # fract. shift by half of the added size
            self.shift_all_atoms(f_shift) # shift without enforcing periodic boundary conditions
        return self.a0 # return new box size
        

    def get_avg_pos(self, l_atoms_idx, proximity, periodic):
        pos = np.array([0.,0.])
        npos = 0
        assert isinstance(l_atoms_idx, list), 'This expects that parameter l_atoms_idx is a list of integers'
        m = len(self.l_atoms) # current number of atoms
        if m == 0: return pos # dummy
        n = len(l_atoms_idx) # number of atoms to include
        if n == 0: return pos # dummy
        mb0 = self.get_basis().T # get the transformation matrix to transform from fractional to physical coordinates
        sdthr = proximity * proximity
        pos = self.l_atoms[l_atoms_idx[0]].pos.copy()
        npos = 1
        if n > 1:
            for i in range(1, n):
                idx = l_atoms_idx[i]
                apos = self.l_atoms[idx].pos.copy()
                dpos = apos - pos
                wpos = np.array([0.,0.,0.])
                if periodic:
                    for j in range(0,3):
                        if dpos[j] < -0.5: wpos[j] = 1.
                        if dpos[j] >= 0.5: wpos[j] = -1.
                bpos = apos + wpos
                dpos = np.dot(mb0, bpos - pos)
                if np.dot(dpos,dpos) < sdthr:
                    pos = (pos * npos + bpos) / (npos+1)
                    npos += 1
        if periodic: return np.round(pos % 1.0, 6)
        return np.round(pos, 6)

    def get_type_dict(self, l_atoms_idx=None, l_type_name_adds=None):
        """

            Returns a dictionary listing atomic types and sites assigned
            to these types for all atoms indexed in list l_atoms_idx.

            Parameters
            ----------
                l_atoms_idx : list
                    List of indices of the structures atom list to be included.
                    Can be None to use all atoms of the structure.
                l_type_name_adds : list
                    List of strings that determine additions made to the
                    type names:
                    'occ' : adds occupancy
                    'uiso' : adds the thermal vibration mean square amplitude
                    'ion' : adds the ionic charge
        """
        if l_atoms_idx is None:
            aidx = list(range(0, len(self.l_atoms)))
        else:
            aidx = l_atoms_idx
        if l_type_name_adds is None:
            li_type_name_adds = []
        else:
            li_type_name_adds = l_type_name_adds
        d = {}
        n = len(aidx)
        m = len(self.l_atoms)
        if (m > 0 and n > 0):
            for i in aidx:
                if (i < m and i >= 0):
                    a = self.l_atoms[i]
                    s = a.get_type_name(li_type_name_adds)
                    if s in d.keys(): # only add site
                        d[s]['sites'].append(a.pos)
                    else: # add new type
                        d[s] = { 'Z' : a.Z, 'occ' : a.occ, 'uiso' : a.uiso, 'ion' : a.charge, 'sites' : [a.pos] }
        return d

    def add_atom(self, Z, uiso, pos, occ = 1.0, charge = 0.0):
        """

        Adds an atom to the structure with given parameters.

        Parameters
        ----------
            Z : int
                atomic number
            uiso : float
                mean squared vibration amplitude in [Ang^2]
            pos : numpy.ndarray([x, y, z], dtype=float)
                fractional coordinates in a supercell
            occ : float, default 1.0
                fractional site occupation
            charge : float, default 0.0
                ionic charge in units of the elementary charge
        
        Returns
        -------
            int
                The index at which the atom was added to the list member l_atoms.
        """
        ato = atom(Z, pos, uiso, occ, charge)
        self.l_atoms.append(ato)
        return len(self.l_atoms)

    def keep_atoms(self, l_atoms_idx):
        """

        Keeps atoms in the supercell that are listed by index in the
        parameter l_atoms_idx. This modifies the list of atoms of the
        supercell object.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                to be kept. Other atoms will be removed.

        Returns
        -------

            int
                Number of remaining atoms in the supercell object

        """
        assert isinstance(l_atoms_idx, list), 'This expects that parameter l_atoms_idx is a list of integers'
        m = len(self.l_atoms) # current number of atoms
        if m == 0: return 0 # nothing to keep
        n = len(l_atoms_idx) # new number of atoms
        if n > 0: # check the list
            l_work = [] # internal list
            for i in l_atoms_idx: # copy only valid indices 
                if (i >= 0) and (i < m):
                    l_work.append(i)
            n = len(l_work)
            l_work = sorted(l_work) # get index list sorted
        #
        if n == 0: # nothing to keep
            self.l_atoms.clear() # erase all
        else:
            l_atoms_cp = deepcopy(self.l_atoms) # get a copy of the atoms list
            self.l_atoms.clear() # clear the current list of atoms
            for i in l_work: # copy atoms back
                self.l_atoms.append(l_atoms_cp[i])
        return len(self.l_atoms)

    def delete_atoms(self, l_atoms_idx):
        """

        Deletes atoms from the supercell that are listed by index in the
        parameter l_atoms_idx. This modifies the list of atoms of the
        supercell object.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                to be deleted. Other atoms will be kept.

        Returns
        -------

            int
                Number of remaining atoms in the supercell object

        """
        assert isinstance(l_atoms_idx, list), 'This expects that parameter l_atoms_idx is a list of integers'
        m = len(self.l_atoms) # current number of atoms
        if m == 0: return len(self.l_atoms) # nothing to delete
        n = len(l_atoms_idx) # number of atoms to delete
        if n > 0: # check the list
            l_work = [] # internal list
            for i in l_atoms_idx: # copy only valid indices 
                if (i >= 0) and (i < m):
                    l_work.append(i)
            n = len(l_work)
            l_work = sorted(l_work, reverse=True) # get index list sorted, reverse
        #
        if n == 0: return len(self.l_atoms) # nothing to delete
        for i in l_work: # copy atoms back
            del self.l_atoms[i]
        return len(self.l_atoms)

    def periodic(self):
        """

        Wraps all atoms periodically back to the cell so their
        fractional coordinates are >=0 and <1.

        """
        if len(self.l_atoms) > 0:
            for at in self.l_atoms:
                p = at.pos
                at.pos = np.round( p % 1.0, 6 ) # wrap with precision of 6 digits

    def set_uiso(self, l_atoms_idx, uiso):
        """

        Sets the uiso parameter of all atoms indexed by list
            l_atoms_idx to the given value.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                for which the uiso parameter is set.

            uiso : float
                Isotropic mean square amplitude of thermal vibrations
                in Angström**2 units.

        Returns
        -------
            
            int
                Number of atoms for which the uiso value was set.

        """
        assert isinstance(l_atoms_idx, list), 'This expects that parameter l_atoms_idx is a list of integers'
        m = len(self.l_atoms) # current number of atoms
        if m == 0: return len(self.l_atoms) # nothing to delete
        n = len(l_atoms_idx) # number of atoms to delete
        l = 0
        if n > 0: # work through the list
            for i in l_atoms_idx: # copy only valid indices 
                if (i >= 0) and (i < m):
                    self.l_atoms[i].uiso = uiso
                    l += 1
        return l

    def set_biso(self, l_atoms_idx, biso):
        """

        Sets the uiso parameter of all atoms indexed by list
            l_atoms_idx by translating the given biso value.
            uiso = biso / (8 Pi**2)

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                for which the biso parameter is set.

            biso : float
                Isotropic B parameter of the Debye-Waller factor
                in Angström**2 units. This relates to the iostropic
                mean squared displacement amplitude usio as
                biso = 8 * Pi**2 * usio.

        Returns
        -------
            
            int
                Number of atoms for which the uiso value was set.

        """
        uiso = biso / (8. * np.pi**2) # from biso to uiso
        return self.set_uiso(l_atoms_idx, uiso)

    def set_occ(self, l_atoms_idx, occ):
        """

        Sets the occupancy parameter of all atoms indexed by list
            l_atoms_idx to the given value.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                for which the occupancy parameter is set.

            occ : float
                occupancy factor, clipped between 0 and 1

        Returns
        -------
            
            int
                Number of atoms for which the occupancy value was set.

        """
        assert isinstance(l_atoms_idx, list), 'This expects that parameter l_atoms_idx is a list of integers'
        m = len(self.l_atoms) # current number of atoms
        if m == 0: return len(self.l_atoms) # nothing to delete
        n = len(l_atoms_idx) # number of atoms to delete
        l = 0
        focc = min(1., max(0., occ))
        if n > 0: # work through the list
            for i in l_atoms_idx: # copy only valid indices 
                if (i >= 0) and (i < m):
                    self.l_atoms[i].occ = focc
                    l += 1
        return l

    def shift_atoms(self, l_atoms_idx, shift, periodic=False):
        """

        Shifts all atoms indexed in list l_atoms_idx by a shift vector
        in fractional coordinates.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                to be shifted.

            shift : numpy ndarray((3),float)
                Shift vector in fractional coordinates

            periodic : boolean
                Flags that periodic wrap should be applied after shifting.

        Returns
        -------

            int
                Number of atoms for which were shifted.

        """
        assert isinstance(l_atoms_idx, list), 'This expects that parameter l_atoms_idx is a list of integers'
        m = len(self.l_atoms) # current number of atoms
        n = len(l_atoms_idx) # number of atoms to delete
        l = 0
        if (n > 0) and (m > 0): # work through the list
            for i in l_atoms_idx: # copy only valid indices 
                if (i >= 0) and (i < m):
                    p = self.l_atoms[i].pos + shift
                    if periodic:
                        self.l_atoms[i].pos = np.round(p % 1.0, 6)
                    else:
                        self.l_atoms[i].pos = p
                    l += 1
        return l

    def shift_all_atoms(self, shift, periodic=False):
        """

        Shifts all atoms by a shift vector in fractional coordinates.

        Parameters
        ----------

            shift : numpy ndarray((3),float)
                Shift vector in fractional coordinates

            periodic : boolean
                Flags that periodic wrap should be applied after shifting.

        Returns
        -------

            int
                Number of atoms for which were shifted.
        
        """
        m = len(self.l_atoms) # current number of atoms
        l = 0
        if m > 0: # work through the list
            for i in range(0, m): # copy only valid indices 
                p = self.l_atoms[i].pos + shift
                if periodic:
                    self.l_atoms[i].pos = np.round(p % 1.0, 6)
                else:
                    self.l_atoms[i].pos = p
                l += 1
        return l

    def shift_atoms_to(self, l_atoms_idx, pos, fraction=1., confinement=np.array([0.,0.,0.]), mode=3):
        """

        Shifts atoms indexed in list l_atoms_idx towards pos by
        a given fraction of the initial distance. Shifts can be confined
        to be parallel to planes or lines.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                to shifted.

            pos : numpy ndarray((3),float)
                Target position in fractional coordinates.

            fraction : float, default 1.
                Fraction of the initial distance to shift.
                Depending on parameter <mode>, the initial distance is
                1: to a plane through <pos> and with normal <confinement>
                2: to a line through <pos> and along direction <confinement>
                3: to the point <pos> without any confinements. 

            confinement : numpy ndarray((3),float), default [0.,0.,0.]
                Vector defining the orientation of a target plane (<mode> == 1)
                or a target line (<mode> == 2). If confinement is a zero vector
                a fallback to <mode> == 3 is used.

            mode : int, default 3
                Mode of confined shifts.
                1 : confinement to a direction perpendicular to a plane.
                2 : confinement to a direction perpendicular to a line.
                3 (and any other): no confinement

        Returns
        -------

            int
                Number of shifted atoms.

        """
        eps = 1.e-7 # small distance threshold
        assert isinstance(l_atoms_idx, list), 'Input <l_atoms_idx> should be a list of numbers.'
        assert isinstance(pos,np.ndarray), 'Input <pos> should be a numpy.ndarry object.'
        assert len(pos.flatten()) == 3, 'Input <pos> should contain 3 numberst.'
        assert type(fraction) is float, 'Input <fraction> should be of float type.'
        assert type(mode) is int, 'Input <mode> should be an integer number.'
        assert isinstance(confinement,np.ndarray), 'Input <confinement> should be a numpy.ndarry object.'
        assert len(confinement.flatten()) == 3, 'Input <confinement> should contain 3 numbers.'
        imode = mode
        l_conf = np.sqrt(np.dot(confinement,confinement)) # get confinement vector length
        if l_conf < eps: # zero confinement vector -> mode fallback to 3
            imode = 3
        else: # finite confinement vector
            n_conf = np.round(confinement / l_conf, 6) # confinement normal vector rounded to 6 digits
        if (imode < 1) or (imode > 3): imode = 3 # internal mode switch limited to 1, 2, or 3
        n = len(l_atoms_idx) # number of atoms to be moved
        m = len(self.l_atoms) # number of atoms in the supercell
        l = 0 # number of shifted atoms
        if (n > 0) and (m > 0): # try moving atoms
            for i in l_atoms_idx: # current atom index
                if (i < 0) or (i >= m): continue # skip invalid atom index
                ati = self.l_atoms[i] # current atom
                vec_d_pos = pos - ati.pos # get the vector from atom to pos
                if mode == 1: # ... to plane distance
                    vec_d = n_conf * np.dot(n_conf, vec_d_pos) # perpendicular vector from atom position to plane
                elif mode == 2: # ... to a line
                    vec_l = n_conf * np.dot(n_conf, vec_d_pos) # component of the distance vector parallel to the line
                    vec_d = vec_d_pos - vec_l # component of the distance vector perpendicular to the line
                else: # ... to pos
                    vec_d = vec_d_pos
                vec_shift = vec_d * fraction # shift vector
                p = np.round(ati.pos + vec_shift, 6) # shifted position
                ati.pos[:] = p[:]
                l += 1
        return l

    def merge(self, other_cell):
        """

        Merges atom list of <other_cell> to this object.

        Parameters
        ----------

            other_cell : emilys.structure.supercell.supercell
                Another supercell object to be merged into this object.

        Returns
        -------

            int
                Number of atoms merged to this object.

        Remarks
        -------

            This requires that <other_cell> has the same size and angles.

        """
        eps = 1.E-6
        assert isinstance(other_cell, supercell), 'This requires that <other_cell> is also a supercell object.'
        assert abs((self.a0[0] - other_cell.a0[0])/self.a0[0]) < eps, 'This requires equal box size (conflict with a0[0]).'
        assert abs((self.a0[1] - other_cell.a0[1])/self.a0[1]) < eps, 'This requires equal box size (conflict with a0[1]).'
        assert abs((self.a0[2] - other_cell.a0[2])/self.a0[2]) < eps, 'This requires equal box size (conflict with a0[2]).'
        assert abs((self.angles[0] - other_cell.angles[0])/self.angles[0]) < eps, 'This requires equal box angles (conflict with angles[0]).'
        assert abs((self.angles[1] - other_cell.angles[1])/self.angles[1]) < eps, 'This requires equal box angles (conflict with angles[1]).'
        assert abs((self.angles[2] - other_cell.angles[2])/self.angles[2]) < eps, 'This requires equal box angles (conflict with angles[2]).'
        m = len(other_cell.l_atoms)
        if m > 0:
            self.l_atoms.extend(other_cell.l_atoms)
        return m

    def insert(self, other_cell, proximity):
        """

        Puts atoms from <other_cell> into this object. First removes all atoms of
        this object that would be closer than proximity to the new atoms.

        Parameters
        ----------

            other_cell : emilys.structure.supercell.supercell
                Another supercell object to be inserted into this object.

            proximity : float
                Minimum distance allowed of existing atoms to the new atoms.
                Any atom in the previous list that is closer than proximity
                to any of the new atoms will be removed before adding the new
                atoms.

        Returns
        -------

            int
                New number of atoms in this object.

        Remarks
        -------

            This requires that <other_cell> has the same size and angles.

        """
        eps = 1.E-6
        assert isinstance(other_cell, supercell), 'This requires that <other_cell> is also a supercell object.'
        assert abs((self.a0[0] - other_cell.a0[0])/self.a0[0]) < eps, 'This requires equal box size (conflict with a0[0]).'
        assert abs((self.a0[1] - other_cell.a0[1])/self.a0[1]) < eps, 'This requires equal box size (conflict with a0[1]).'
        assert abs((self.a0[2] - other_cell.a0[2])/self.a0[2]) < eps, 'This requires equal box size (conflict with a0[2]).'
        assert abs((self.angles[0] - other_cell.angles[0])/self.angles[0]) < eps, 'This requires equal box angles (conflict with angles[0]).'
        assert abs((self.angles[1] - other_cell.angles[1])/self.angles[1]) < eps, 'This requires equal box angles (conflict with angles[1]).'
        assert abs((self.angles[2] - other_cell.angles[2])/self.angles[2]) < eps, 'This requires equal box angles (conflict with angles[2]).'
        m = len(other_cell.l_atoms)
        r_check = proximity * 1.1 # make the check range a bit larger than the proximity
        fr_check = r_check * np.reciprocal(self.a0) # fractional check range radius
        if m > 0:
            for atj in other_cell.l_atoms:
                f0 = atj.pos - fr_check # lower check box bounds
                f1 = atj.pos - fr_check # upper check box bounds
                l_rng = self.list_atoms_in_range({'rng_pos_a' : [f0[0], f1[0]], 'rng_pos_b' : [f0[1], f1[1]], 'rng_pos_c' : [f0[2], f1[2]]})
                l_rem = self.remove_close_atoms(l_rng, proximity)
                if len(l_rem) > 0: self.delete_atoms(l_rem) # delete
            self.l_atoms.extend(other_cell.l_atoms) # extend atom list by atoms of other cell
        return len(self.l_atoms)

    def list_positions(self, l_atoms_idx):
        """

        Returns a list of positions of atoms identified by index.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                for which positions should be listed.

        Returns
        -------

            list
                List of atom positions

        """
        assert isinstance(l_atoms_idx, list), 'Input <l_atoms_idx> should be a list of numbers.'
        l_pos = []
        n = len(l_atoms_idx) # number of atoms to be moved
        m = len(self.l_atoms) # number of atoms in the supercell
        if (m > 0) and (n > 0):
            for i in l_atoms_idx:
                if (i < 0) or (i >= m): continue # invalid index
                l_pos.append(self.l_atoms[i].pos)
        return l_pos

    def list_close_atoms(self, l_atoms_idx, proximity, periodic=True, debug=False):
        """

        Returns a list of lists of atoms, which are closer than the
        proximity parameter in nanometers. The periodic option switches
        the check of proximity under periodic boundary conditions.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                to be checked for mutual proximity. Atoms not included
                in the list will be ignored in the proximity checks.

            proximity : float
                Sets a threshold to which distance in nanometers is
                identified as close.

            periodic : boolean, default: True
                Switches proximity checks under periodic boundary
                condistions.

            debug : boolean, default: False
                Switches extra debug text output.

        Returns
        -------

            list
                List of lists of atom indices, each sub-list is a set of atoms
                that are closer to each other than proximity
        
        """
        l_close = []
        assert isinstance(l_atoms_idx, list), 'Input <l_atoms_idx> should be a list of numbers.'
        m = len(self.l_atoms) # number of atoms in the supercell
        n = len(l_atoms_idx) # list of atom indices to check for proximity
        mb0 = self.get_basis().T # get the transformation matrix to transform from fractional to physical coordinates
        sdthr = proximity * proximity
        if (n > 1) and (m > 1): # need at least two atoms to check
            for i in range(0, n-1): # loop over atoms in list, exclusive the last
                idx = l_atoms_idx[i]
                vlp0 = self.l_atoms[idx].pos
                l_close_cur = [idx]
                for j in range(i+1, n): # loop over atoms behind i in the list
                    jdx = l_atoms_idx[j]
                    vlp1 = self.l_atoms[jdx].pos
                    if periodic: # fractional distance vector across periodic boundary conditions
                        vdlp = ((vlp1 - vlp0 + 0.5) % 1.0 ) - 0.5
                    else: # fractional distance vector, no periodic boundary
                        vdlp = vlp1 - vlp0
                    vdp = np.dot(mb0, vdlp) # distance vector in physical coordinates [nm]
                    sd = np.dot(vdp, vdp)
                    if sd <= sdthr: # squared distance check [nm**2]
                        if debug: print('#{:d} {:s} <-> #{:d} {:s}: d = {:.4f} nm'.format(
                            idx, aty.atom_type_symbol[self.l_atoms[idx].Z],
                            jdx, aty.atom_type_symbol[self.l_atoms[jdx].Z], np.sqrt(sd)))
                        l_close_cur.append(jdx) # add to current list
                # handle the current list of atoms close to atim idx
                if len(l_close_cur) > 1: # at least a pair?
                    l_close.append(l_close_cur) # append to output list
                    if debug: print('added list', l_close_cur)
        return l_close

    def remove_close_atoms(self, l_atoms_idx, proximity, debug=False):
        """

        Returns a list of atom indices to be removed from l_atoms_idx. The removel is
        not performed, so that l_atoms_idx remains unchanged by this routine.
        The list is parsed in sequence, checking proximity between atom i and atom i + x.
        Double checking should not occur in this implementation.

        Parameters
        ----------

            l_atoms_idx : list
                List of indices identifying atoms in member l_atoms
                to be checked for mutual proximity. Atoms not included
                in the list will be ignored in the proximity checks.

            proximity : float
                Sets a threshold to which distance in nanometers is
                identified as close.

            debug : boolean, default: False
                Switches extra debug text output.

        Returns
        -------

            list
                List of atom indices that are too close to one other member of
                the input list. Removing those indices will leave a list where no
                atom is closer than proximity to any other.

        """
        l_close = []
        if debug: print('remove_close_atoms:')
        assert isinstance(l_atoms_idx, list), 'Input <l_atoms_idx> should be a list of numbers.'
        l_in = l_atoms_idx.copy() # make a copy to work with
        n = len(l_in) # list of atom indices to check for proximity
        mb0 = self.get_basis().T # get the transformation matrix to transform from fractional to physical coordinates
        sdthr = proximity * proximity
        if debug: print('- number of input items:', n)
        if (n > 1): # there are at least two atoms
            for i in range(0, n-1): # loop over atoms in list, exclusive the last
                idx = l_in[i]
                vlp0 = self.l_atoms[idx].pos
                l_rem = []
                for j in range(i+1, n): # loop over atoms behind i in the list
                    jdx = l_in[j]
                    vlp1 = self.l_atoms[jdx].pos
                    vdlp = vlp1 - vlp0 # fractional distance vector, no periodic boundary
                    vdp = np.dot(mb0, vdlp) # distance vector in physical coordinates [nm]
                    sd = np.dot(vdp, vdp) # square distance
                    if sd <= sdthr: # squared distance check [nm**2]
                        l_rem.append(j) # add local index to current removal list
                        l_close.append(jdx) # add atom index to output removal list
                # handle the current list of close atoms
                if len(l_rem) > 0: # at least one atom is too close?
                    if debug: print('- (', i, ') removing', len(l_rem), 'items ...')
                    l_rem.reverse() # reverses the local index list, so to delete from the end downwards
                    for j in l_rem: # loop local indices to remove from l_in
                        del l_in[j] # delete from l_in
                    n = len(l_in) # update lenth of l_lin
                    if debug: print('- (', i, ') remaining items:', n)
                if i >= n - 1: break # stop here, list has become too short
        return l_close

    def list_atoms_in_range(self, dic_range={}):
        """

        Returns a list of indices of atoms in member l_atoms
        whose parameters are within in all of the ranges defined
        in the dictionary dic_range.

        This is an alternative implementation

        Parameters
        ----------

            dic_range : dict, default {}
                dictionary of range definitions
                supported range keys are
                'rng_Z' : [int, int]
                    atomic numbers
                'rng_charge' : [float, float]
                    ionic charges
                'rng_pos_a' : [float, float]
                    fractional atom position along cell a axis
                'rng_pos_b' : [float, float]
                    fractional atom position along cell b axis
                'rng_pos_c' : [float, float]
                    fractional atom position along cell c axis
                'rng_uiso' : [float, float]
                    thermal vibration amplitudes
                'rng_occ' : [float, float]
                    occupancy factors

        Returns
        -------

            list of int

        Notes
        -----

            Ranges defined will be checked inclusive of the lower
            bound and exclusive for the upper bound,
            i.e. (x0 <= x) and (x < x1).

        """
        l_atoms_idx = [] # initialize empty list of selected atom indices
        n_atoms = len(self.l_atoms) # get number of atoms in the structure
        if n_atoms > 0:
            for i in range(0, n_atoms): # loop over all atoms
                ati = self.l_atoms[i]
                b_add = True # assume adding and perform logical AND operations to turn it off in case of failing conditions
                for sel_key in dic_range: # go through all conditions
                    min_val = min(dic_range[sel_key])
                    max_val = max(dic_range[sel_key])  
                    if sel_key == 'rng_Z': # atomic number range condition
                        b_add &= ((ati.Z >= min_val) and (ati.Z < max_val))
                    if sel_key == 'rng_charge': # charge range
                        b_add &= ((ati.charge >= min_val) and (ati.charge < max_val))
                    if sel_key == 'rng_uiso': # uiso range
                        b_add &= ((ati.uiso >= min_val) and (ati.uiso < max_val))
                    if sel_key == 'rng_occ': # occupancy range
                        b_add &= ((ati.occ >= min_val) and (ati.occ < max_val))
                    if sel_key == 'rng_pos_a': # position x range
                        b_add &= ((ati.pos[0] >= min_val) and (ati.pos[0] < max_val))
                    if sel_key == 'rng_pos_b': # position y range
                        b_add &= ((ati.pos[1] >= min_val) and (ati.pos[1] < max_val))
                    if sel_key == 'rng_pos_c': # position z range
                        b_add &= ((ati.pos[2] >= min_val) and (ati.pos[2] < max_val))
                if b_add: # all conditions fullfilled ...
                    l_atoms_idx.append(i) # add atom index to list
        return l_atoms_idx


# -*- coding: utf-8 -*-
"""
Created on Mon May 03 08:46:00 2021
@author: ju-bar

Functions handling input and output of structure data
via the CEL file format.

The CEL file format is a text format defining a volume in space
which is filled by atoms.

The supercell is defined by a sequence of numbers in the second
line of the file. The line starts with a '0' characters, followed
by three numbers which set the lengths a, b, and c of the cell
edges in nanometers and three more number setting the angles
alpha, beta, and gamma  in degrees between these edges.

Atoms are defined by further lines of text, with one line for
each atom in the cell. The element is defined by symbol and the
atomic position in fractional coordinates of the cell. An occupancy
factor between 0 and 1 must be given as well as a temperature factor
using the B parameter of Debye Waller factors in nanometer square
units. Three more numbers are expected at the end of each atom line
but these numbers have no effect.

A '*' character should be placed in the last line to signalize the
end of structure input.

Example:
# This is the first line of the example defining a Si crystal in [110] orientation
  0   0.38400  0.54305  0.38400 90.00000 90.00000 90.00000
   Si  0.000000 0.000000 0.250000 1.000000 0.005400 0.000000 0.000000 0.000000
   Si  0.000000 0.250000 0.750000 1.000000 0.005400 0.000000 0.000000 0.000000
   Si  0.500000 0.750000 0.250000 1.000000 0.005400 0.000000 0.000000 0.000000
   Si  0.500000 0.500000 0.750000 1.000000 0.005400 0.000000 0.000000 0.000000
*

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import re
import numpy as np
from emilys.structure.supercell import supercell
import emilys.structure.atom as ato
import emilys.structure.atomtype as aty

def get_atom_str_CEL(a):
    """
    Returns a string of atom object parameters in CEL file format.

    Parameters
    ----------
        
        a : atom
            atom object

    Returns
    -------
        string

    """
    assert isinstance(a, ato.atom), 'This expects input of type emilys.structure.atom.atom'
    symb = aty.atom_type_symbol[a.Z]
    biso = 8. * np.pi**2 * a.uiso * 0.01 # biso in nm^2
    scrg = ato.get_str_from_charge(a.charge)
    s_out = ' {:s} {:>10.6f} {:>10.6f} {:>10.6f} {:>10.6f} {:>10.6f}'.format(
            symb + scrg, a.pos[0], a.pos[1], a.pos[2], a.occ, biso)
    return s_out + '  0.000000  0.000000  0.000000'

def set_atom_str_CEL(s):
    """
    Returns an atom object with parameters set from input string s.

    Parameters
    ----------
        
        s : string
            CEL file line defining an atom in a super cell
            '<symbol> <frac-x> <frac-x> <frac-z> <occ> <biso nm^2> 0 0 0'

    Returns
    -------
        atom

    """
    assert len(s) >= 11, 'insufficient length of input string'
    assert isinstance(s, str), "input should be a str object"
    l_cmp = re.split(' +|,|;|\t+', s.strip()) # decompose to list of string
    symb, in_charge = ato.get_symb_charge(l_cmp[0])
    in_Z = aty.Z_from_symb(symb) # atom type
    in_pos = np.array([float(l_cmp[1]), float(l_cmp[2]), float(l_cmp[3])])
    in_occ = float(l_cmp[4])
    in_uiso = float(l_cmp[5]) * 100. / (8. * np.pi**2) #  usio in Angst^2
    return ato.atom(Z=in_Z, pos = in_pos, uiso=in_uiso, occ=in_occ, charge=in_charge)

def read_CEL(file, debug=False):
    """
    
    Reads atomic structure data from a file assuming the CEL file format.

    Parameters
    ----------

        file : string
            Input file name.
        debug : boolean, default: False
            Flags debug print out of the file parsing.

    Returns
    -------

        emilys.structure.supercell.supercell

    """
    assert type(file) is str, "This expects a string as input parameter."
    # open the file and read the lines
    with open(file) as file_in:
        if debug: print('dbg (read_cel): opened file [' + file + ']')
        lines = []
        for line in file_in:
            lines.append(line)
        file_in.close()
    if debug: print('dbg (read_cel): read ', len(lines),' lines of text.')
    assert len(lines) > 2, "The input file doesn't contain sufficient number of text lines."
    # create a supercell object
    sc = supercell()
    # process the cell parameters
    l_cmp = re.split(' +|,|;|\t+', lines[1].strip()) # decompose to list of string
    assert len(l_cmp) > 6, "The second line of the input file could not be split into >6 items."
    if debug: print('dbg (read_cel): supercell input line: ', l_cmp)
    sc.a0 = np.array([float(l_cmp[1]),float(l_cmp[2]),float(l_cmp[3])]) * 10. # from nm to Angst
    if debug: print('dbg (read_cel): supercell size [A]: a = {:.5f}, a = {:.5f}, c = {:.5f}'.format(*sc.a0))
    sc.angles = np.array([float(l_cmp[4]),float(l_cmp[5]),float(l_cmp[6])])
    if debug: print('dbg (read_cel): supercell angles [deg]: alpha = {:.4f}, beta = {:.4f}, gamma = {:.4f}'.format(*sc.angles))
    # process the lines
    for i in range(2, len(lines)):
        if '*' in lines[i]: break # stop reading due to end of structure character
        at_in = set_atom_str_CEL(lines[i])
        sc.l_atoms.append(at_in)
    if debug: print('dbg (read_cel): added ', len(sc.l_atoms), ' atoms to the supercell.')
    return sc

def write_CEL(sc, file):
    """

    Writes atomic structure data in CEL file format.

    Parameters
    ----------

        sc : emilys.structure.supercell.supercell
            Supercell data the should be written to file.
        file : str
            File name
    
    Returns
    -------

        int
            Error code

    """
    io_err = 0
    assert isinstance(sc, supercell), 'This expects that sc is input of type emilys.structure.supercell.supercell'
    assert isinstance(file, str), 'This expects that file is input of type str'
    with open(file, "w") as file_out:
        file_out.write("# EMILYS celio of [" + sc.get_composition_str() + "] \n")
        lat_c = sc.a0 * 0.1 # from Angst to nm
        file_out.write(" 0  {:.5f}  {:.5f}  {:.5f}  {:.4f}  {:.4f}  {:.4f} \n".format(
            lat_c[0], lat_c[1], lat_c[2], sc.angles[0], sc.angles[1], sc.angles[2]))
        for a in sc.l_atoms:
            file_out.write(get_atom_str_CEL(a) + "\n")
        file_out.write("*\n")
        file_out.close()
    return io_err


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 30 10:52:00 2021
@author: ju-bar

Functions and classes handling atom data

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import re
import numpy as np
import emilys.structure.atomtype as atty

def get_str_from_charge(charge, digits=2):
    """
    Returns a string to be attached to a symbol for a given charge.
    """
    sgn = ''
    acrg = abs(charge)
    sacrg = '{0:.{1}f}'.format(acrg,digits)
    lsa = len(sacrg)
    s1 = sacrg[:lsa-digits-1]
    s2 = sacrg[lsa-digits:]
    if int(s1)>0 or int(s2)>0:
        sgn = '+'
        if charge < 0.: sgn = '-'
    if int(s2) > 0:
        s2 = '.' + s2.rstrip('0')
    else:
        s2 = ''
    sout = ''
    if int(s1) > 0 or len(s2) > 0:
        sout = s1 + s2 + sgn
    return sout

def get_symb_charge(s):
    """
    Returns symbol prefix and charge number from string s
    """
    charge = 0.
    symbol = ''
    pn = 0
    m = re.search('[0-9]+', s)
    if m:
        pn = m.start()
        symbol = s[0:pn]
        sgn = 1.
        if s[len(s)-1] == '-': sgn = -1.
        charge = float(s[pn:len(s)-1]) * sgn
    else:
        symbol = s
    return symbol, charge

        

class atom:
    """

    class atom

    Parameters
    ----------

        Z : integer
            atomic number
        charge : float
            ionic charge in units of the elementary charge
        uiso : float
            mean squared vibration amplitude in [Ang^2]
        occ : float
            fractional site occupation
        pos : numpy.ndarray([x, y, z], dtype=float)
            fractional coordinates in a supercell

    Methods
    -------

        get_type_name(l_type_name_adds)
            Returns a string for an atom type name with possible addistions
            depending on other atom parameters.

    """

    def __init__(self, Z=1, pos=np.array([0.,0.,0.]), uiso=0.006332574, occ=1., charge=0.):
        self.Z = Z
        self.pos = pos
        self.uiso = uiso
        self.occ = occ
        self.charge = charge

    def get_type_name(self, l_type_name_adds = []):
        """
        
        Returns a string for an atom type name with possible additions
        depending on other atom parameters.
        
        Parameters
        ----------

            l_type_name_adds : list
                List of strings that determine additions made to the
                type names:
                'occ' : adds occupancy
                'uiso' : adds the thermal vibration mean square amplitude
                'ion' : adds the ionic charge

        """
        s = atty.atom_type_symbol[self.Z]
        scrg = ''
        socc = ''
        suiso = ''
        for sadd in l_type_name_adds:
            if sadd == 'ion':
                scrg = get_str_from_charge(self.charge) 
            if sadd =='occ' :
                socc = '_occ{:.3f}'.format(self.occ)
            if sadd == 'uiso':
                suiso = '_uiso{:.6f}'.format(self.uiso)
        return s + scrg + socc + suiso

    



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 30 10:34:00 2021
@author: ju-bar

Functions and classes handling atom type data

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

atom_type_symbol = ['Void'  , 
             'H'	, 'He'	, 'Li'	, 'Be'	, 'B'	, 'C'	, 'N'	, 'O'	, 'F'	, 'Ne'	,
			 'Na'	, 'Mg'	, 'Al'	, 'Si'	, 'P'	, 'S'	, 'Cl'	, 'Ar'	, 'K'	, 'Ca'	,
			 'Sc'	, 'Ti'	, 'V'	, 'Cr'	, 'Mn'	, 'Fe'	, 'Co'	, 'Ni'	, 'Cu'	, 'Zn'	,
			 'Ga'	, 'Ge'	, 'As'	, 'Se'	, 'Br'	, 'Kr'	, 'Rb'	, 'Sr'	, 'Y'	, 'Zr'	,
			 'Nb'	, 'Mo'	, 'Tc'	, 'Ru'	, 'Rh'	, 'Pd'	, 'Ag'	, 'Cd'	, 'In'	, 'Sn'	,
			 'Sb'	, 'Te'	, 'I'	, 'Xe'	, 'Cs'	, 'Ba'	, 'La'	, 'Ce'	, 'Pr'	, 'Nd'	,
			 'Pm'	, 'Sm'	, 'Eu'	, 'Gd'	, 'Tb'	, 'Dy'	, 'Ho'	, 'Er'	, 'Tm'	, 'Yb'	,
			 'Lu'	, 'Hf'	, 'Ta'	, 'W'	, 'Re'	, 'Os'	, 'Ir'	, 'Pt'	, 'Au'	, 'Hg'	,
			 'Tl'	, 'Pb'	, 'Bi'	, 'Po'	, 'At'	, 'Rn'	, 'Fr'	, 'Ra'	, 'Ac'	, 'Th'	,
			 'Pa'	, 'U'	, 'Np'	, 'Pu'	, 'Am'	, 'Cm'	, 'Bk'	, 'Cf'	, 'Es'	, 'Fm'	,
			 'Md'	, 'No'	, 'Lr'	, 'Rf'	, 'Db'	, 'Sg'	, 'Bh'	, 'Hs'	, 'Mt'	, 'Ds'	,
			 'Rg'	, 'Cn'	, 'Nh'	, 'Fl'	, 'Mc'	, 'Lv'	, 'Ts'	, 'Og',]

def Z_from_symb(symb):
    """

    Given atomic symbol, symb, return the atomic number Z.

    Example: Z_from_symb('Ti') -> 22.

    """
    lsy = str(symb)
    lsx = lsy.translate(lsy.maketrans(dict.fromkeys('+-.0123456789','')))
    return [s.upper() for s in atom_type_symbol].index(lsx.upper())






# -*- coding: utf-8 -*-
"""
Created on Wed Sep 14 12:18:00 2022
@author: ju-bar

EMD file I/O

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

import json
import h5py
import numpy as np

def get_data_as_str(data):
    d = np.asanyarray(data)
    string_ = [chr(i[0]) for i in d]
    return ''.join(string_).strip('\x00')

def read_emd(name, debug = False):
    """

    Reads image data from an EMD file saved by the TFS Velox software.

    The data is returned in a dictionary with primary keys given by
    detector names. For each detector, metadata is in the secondary
    key 'Metadata' and the image data is in the secondary key 'Data'.
    A secondary key 'FrameLookupTable' is present.
    The image data is in ordering [nframe, ny, nx].

    Parameters
    ----------

        str : name
            File name
        boolean : debug, default : False
            Switch for debug text output

    Returns
    -------
        dict : data and metadata

    """

    d_out = {}

    # load the file
    if debug: print('emd_read_stem: reading file [' + name + ']')
    d_emd = h5py.File(name, 'r')

    main_keys = ['Version', 'Info']
    for k in main_keys:
        if k in d_emd:
            if debug: print("  reading \'" + k + "\'")
            d_out[k] = json.loads(d_emd[k][0])

    # check for image operations
    if 'Operations' in d_emd: # add operations applying to images
        if debug: print("- file contains operations information")
        d_op = d_emd['Operations']
        d_out['Operations'] = {} # create operations key in output dict
        if 'Operations' in d_op: # operation sequence ...
            if debug: print("  reading operations sequence")
            d_out['Operations']['Operations'] = json.loads(d_op['Operations'][0])
        # set target operation information to transfer also ...
        op_groups = ['CameraInputOperation', 'StemInputOperation', 
                'SurfaceReconstructionOperation', 'MathematicsOperation', 'DisplayLevelsOperation' ,
                'DpcOperation', 'IntegrationOperation', 'FftOperation']
        for k in op_groups:
            if k in d_op:
                d_op_group = d_op[k] # found a target group of operations
                if debug: print("  reading operation group " + k)
                d_out['Operations'][k] = {}
                for op in d_op_group:
                    d_out['Operations'][k][op] = json.loads(d_op_group[op][0])

    # check for image data
    d_img = d_emd['Data/Image'] # get the /Data/Image HDF5 group
    l_img_keys = list(d_img.keys()) # get the group names (cryptic)
    if debug: print('- found', len(l_img_keys), 'image groups')

    for img_group in l_img_keys: # loop over image groups
        if debug: print('- image group = ' + img_group)
        
        dataPath = '/Data/Image/' + img_group
        name_image = img_group # preset image name with the cryptic key, we are trying to replace that now by a more sensible name if possible
        # the more sensible name can be found in ...
        display_name_lookup = '/Presentation/Displays/ImageDisplay/'
        did = {}
        for key in d_emd[display_name_lookup]:
            did = json.loads(d_emd[display_name_lookup + key][0])
            if did['dataPath'] == dataPath:
                name_image = did['display']['label']
                did['operation'] = display_name_lookup + key # insert key for lookup in 'DisplayLevelsOperation'
                if debug: print('  image display name = ' + name_image + ' (used as primary key)')
                break # stop looping over display keys
        
        # generate a group with primary key name
        d_out[name_image] = {}
                
        # add primary data from the emd file for data identification
        d_out[name_image]['dataPath'] = dataPath
        d_out[name_image]['ImageDisplay'] = did

        if debug: print('  reading metadata ...')
        d = json.loads(get_data_as_str(d_img[img_group]['Metadata'])) #  get the metadata

        # store metadata in the dictionary under key = img_group
        d_out[name_image]['Metadata'] = d
        # ...   frame lookup table
        d_out[name_image]['FrameLookupTable'] = np.asanyarray(d_img[img_group]['FrameLookupTable']).reshape(d_img[img_group]['FrameLookupTable'].shape)
        
        if debug: print('  reading image data ...')
        # ...   the image data
        dt = d_img[img_group]['Data'].dtype
        h5data = d_img[img_group]['Data']
        data = np.empty(h5data.shape)
        h5data.read_direct(data)
        data = np.rollaxis(data, axis=2) # Set the axes in frame, y, x order
        d_out[name_image]['Data'] = data.astype(dt)
        if debug: print('    image data shape =', d_out[name_image]['Data'].shape, '  type =', d_out[name_image]['Data'].dtype)
        del h5data
        del data
        

    # close the file
    d_emd.close()

    return d_out


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 28 13:43:00 2022
@author: ju-bar

Differential phase contrast routines

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

import numpy as np

def dpc_image(a, b, c, d, alpha = 0.0):
    """

    Calculates DPC images from for segmented STEM images a, b, c, d
    under rotation alpha in the diffraction plane.

    Parameters
    ----------

        a, b, c, d : numpy.ndarray, (dimension = 2)
            segmented STEM images
        alpha : float
            rotation angle [radians]

    Returns
    -------

        numpy.ndarray (complex)
            real part is the x DPC component
            imaginary part is the y DPC component

    """
    dx = (a - c).astype(np.float64)
    dy = (b - d).astype(np.float64)
    z = np.empty(dx.shape, dtype=np.complex128)
    sa = np.sin(alpha)
    ca = np.cos(alpha)
    z.real = dx * ca - dy * sa
    z.imag = dy * ca + dx * sa
    return z

def idpc_image(a, b, c, d, alpha=0., rk = [0., 1.], fp = 2.):
    """

    Calculates an iDPC image from for segmented STEM images a, b, c, d
    under rotation alpha in the diffraction plane and band-pass filter.

    Parameters
    ----------

        a, b, c, d : numpy.ndarray, (dimension = 2)
            segmented STEM images
        alpha : float, default 0.
            rotation angle [radians]
        rk : list (len=2), default [0., 1.]
            band pass filter threshold in fractions of the
            image Nyquist frequency
        fp : float, default 2.
            frequency filter exponential power
            formula: exp(-1.0 * (k2 / thr)**fp)

    Returns
    -------

        numpy.ndarray (complex)
            real part is the x DPC component
            imaginary part is the y DPC component

    """
    nd = a.shape
    pf = -1.J / (2. * np.pi)
    kx = np.tile([np.fft.fftfreq(nd[1])],(nd[0],1))
    ky = np.tile([np.fft.fftfreq(nd[0])],(nd[1],1)).T
    k2 = kx**2 + ky**2
    k2[0,0] = 1.0
    ik2 = 1.0 / k2
    ik2[0,0] = 0.0
    kmin = max(1. / nd[0], 1. / nd[1])
    ft02 = max(kmin, rk[0])**2
    ft12 = max(kmin, rk[1])**2
    flt = (1. - np.exp(-1.0 * (k2 / ft02)**fp)) * np.exp(-1.0 * (k2 / ft12)**fp)
    z = dpc_image(a, b, c, d, alpha)
    fzx = np.fft.fft2(z.real)
    fzy = np.fft.fft2(z.imag)
    fidpc = pf * (kx * fzx + ky * fzy) * flt * ik2
    idpc = np.fft.ifft2(fidpc)
    #print(np.sum(idpc.real),np.sum(idpc.imag))
    return idpc.real







# -*- coding: utf-8 -*-
"""
Created on Sun Sep 17 17:30:00 2020
@author: ju-bar

Polar image transformation

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
#
import numba # include compilation support
import numpy as np # include numeric functions
import emilys.image.imagedata as aimg # include image access routines
from emilys.numerics.roots import root_poly_2 # include root finder for square equations
from timeit import default_timer as timer
#
def to_polar_2d(x):
    """

    For input tuples x = [x1,x2] calculates radius and azimuth [r,phi]
    with r = (x1**2 + x2**2)**0.5 and phi = arctan2(x2,x1)
    
    Parameters
    ----------
        x : tuple of floats
            cartesian x1, x2 coordinates
    
    Returns
    -------
        rp : tuple of floats
            [r, phi] 

    """
    return np.array([np.sqrt(x[0]**2 + x[1]**2), np.arctan2(x[1], x[0])])
#
def from_polar_2d(r, phi, x, y):
    """

    For input tuples [r,phi] calculates cartesian positions [x1,x2]
    with x1 = r * cos(phi), x2 = r * sin(phi)
    
    Parameters
    ----------
        r : float, input
            polar radius
        phi : float, input
            polar aziumth
        x : float, output
            cartesian x coordinate
        y : float, output
            cartesian y coordinate
    
    """
    x = r * np.cos(phi)
    y = r * np.sin(phi)
#
@numba.jit
def polar_resample_core(img_xy, x0, y0, dx, dy, img_pr, r0, p0, dr, dp, ipol):
    """

    Polar resampling algorithm with simplified interface for the numba JIT compiler.
    
    Parameters
    ----------
        img_xy : numpy.ndarray(shape=(ny,nx),dtype=float)
            input image
        x0 : float
            x-origin of the cartesian coordinates in input grid pixels
        y0 : float
            y-origin of the cartesian coordinates in input grid pixels
        dx : float
            x-sampling rate of the cartesian coordinates in input grid pixels
        dy : float
            y-sampling rate of the cartesian coordinates in input grid pixels
        img_pr : numpy.ndarray(shape=(nr,np),dtype=float)
            output image
        r0 : float
            start radius in from the cartesian origin in scaled units
        p0 : float
            start azimuth from the cartesian x-axis in radians
        dr : float
            step size of the radial axis in the output in scaled units
        dp : float
            step size of the azimuth axis in the output in radians
        ipol : int
            interpolation switch
            0 : nearest neighbor
            1 : bi-linear
    """
    nyx = img_xy.shape
    #print("polar_resamp_core: ndim1:", nyx)
    nrp = img_pr.shape
    #print("polar_resamp_core: ndim2:", nrp)
    # initialize
    x_0 = x0 * dx
    y_0 = y0 * dy
    # loop over polar coordinates
    for ir in range(0, nrp[0]): # loop over radii
        r = r0 + dr * ir # radial coordinate
        for ip in range(0, nrp[1]): # loop over azimuth
            p = p0 + dp * ip # azimuthal coordinate
            ix = (x_0 + r * np.cos(p)) / dx # x pixel position on input grid
            iy = (y_0 + r * np.sin(p)) / dy # y pixel position on input grid
            if ((ix >= 0.) and (ix <= nyx[1]-1.) and (iy >= 0.) and (iy <= nyx[0]-1.)):
                pos = np.array([ix, iy])
                # img_pr[ir,ip] = aimg.image_at(img_xy, pos, ipol) # get from input image by interpolation
                if (ipol == 1): # bilinear interpolation
                    il = int(np.floor(pos[0])) # x
                    ih = il + 1
                    fi = pos[0] - il
                    jl = int(np.floor(pos[1])) # y
                    jh = jl + 1
                    fj = pos[1] - jl
                    v  = img_xy[jl,il] * (1-fj) * (1-fi)
                    v += img_xy[jl,ih] * (1-fj) * fi
                    v += img_xy[jh,il] * fj * (1-fi)
                    v += img_xy[jh,ih] * fj * fi
                    img_pr[ir,ip] = v
                else: # fall-back to nearest neighbor interpolation
                    jx = int(np.round(ix))
                    jy = int(np.round(iy))
                    img_pr[ir,ip] = img_xy[jy,jx]
    return img_pr
#
def polar_resample(image, num_rad, num_phi, pole, rng_rad, rng_phi = [0.,2. * np.pi], image_scale = [1., 1.], ipol = 1):
    """

    Transforms 2D images to polar grids using interpolation.

    Parameters
    ----------
        image : numpy.array of 2 dimensions
            data on a regular grid
        num_rad : integer
            number of radial sample
        num_phi : integer
            number of azimuthal samples
        pole : float array of length 2
            position (x,y) of the pole/origin in the input image (fractional pixel coordinate)
        rng_rad : float array of length 2
            radial range start and coverage (r_0, delta_r)
        rng_phi : float array of length 2
            azimuthal range start and coverage (phi_0, delta_phi) in radians
            default = [0., 2.*np.pi]
        image_scale : float array of length 2
            scale (x, y) of the input image in physical units per pixel, e.g. nm/pixel
            default = [1., 1.]
        ipol : int
            interpolation type
            0 = nearest neigbor
            1 = bi-linear

    Returns
    -------
        numpy.array of 2 dimensions, shape (num_rad,num_phi)

    Remarks
    -------
        * May produce strange signal with noisy data and if the polar sampling
          is too fine compared to the input grid. The latter usually happens
          around the pole.
        * Non-isotropic sampling of the input is optional via the scale argument.
        * If the scale argument is used, the rng_rad input is assumed to be meant
          on the same scale as the input scale, i.e. the same physical scale unit.
          Also the azimuth argument is assumed to address physical angles, which may
          deviate from angles of a regular grid.
        * Due to the non-isotropic sampling of a polar system, the norm of the
          input will not be conserved in the output. If this is important, you
          may want to try a re-binning algorithm such as in polar_rebin below.
        * The routine uses an interpolation wth @jit decorator. This may require
          more time for compiling during the first call.

    """
    nd = image.shape
    assert len(nd)>=2, 'need at least a 2d array'
    ld = len(nd)
    nx = nd[ld-1]
    ny = nd[ld-2]
    nimg = int(image.size / nx / ny)
    img = image.reshape(nimg, ny, nx) # reshape to a sequence of 2d images
    assert num_rad>=1, 'number of radial samples must be at least 1'
    assert num_phi>=1, 'number of azimuthal samples must be at least 1'
    npol = int(np.array(pole).size / 2)
    p_org = np.array(pole).reshape(npol,2) # sequence of pole positions
    assert (npol==nimg or npol==1), 'invalid number of pole positions'
    assert len(rng_rad)==2, 'invalid length of radial range input'
    assert np.abs(rng_rad[1])>0., 'invalid radial range parameter'
    assert len(rng_phi)==2, 'invalid length of azimuthal range input'
    assert np.abs(rng_phi[1])>0., 'invalid azimuth range parameter'
    assert len(image_scale)>=2, 'invalid length of input image scale parameter'
    assert (np.abs(image_scale[0])>0. and np.abs(image_scale[1])>0.), 'invalid input image scale'
    assert (ipol>=0 and ipol<=1), 'interpolation switch {:d} not supported'.format(ipol)
    # initialize
    r_0 = rng_rad[0]
    d_r = rng_rad[1] / num_rad
    p_0 = rng_phi[0]
    d_p = rng_phi[1] / num_phi
    d_x = image_scale[0]
    d_y = image_scale[1]
    image_out = np.zeros((nimg,num_rad,num_phi)) # polar data
    x_0 = p_org[0,0]
    y_0 = p_org[0,1]
    for i in range(nimg):
        if (npol>1):
            x_0 = p_org[i,0]
            y_0 = p_org[i,1]
        polar_resample_core(img[i,:,:], x_0, y_0, d_x, d_y, image_out[i,:,:], r_0, p_0, d_r, d_p,ipol)
    nout = np.array(nd).astype(int)
    nout[ld-2] = num_rad
    nout[ld-1] = num_phi
    return image_out.reshape(nout)
#
#
@numba.jit
def polar_rebin_core(img_xy, x0, y0, dx, dy, img_pr, r0, p0, dr, dp, acc_norm = False):
    """

    Polar rebinning algorithm with simplified interface for the numba JIT compiler.
    
    Parameters
    ----------
        img_xy : numpy.ndarray(shape=(ny,nx),dtype=float)
            input image
        x0 : float
            x-origin of the cartesian coordinates in input grid pixels
        y0 : float
            y-origin of the cartesian coordinates in input grid pixels
        dx : float
            x-sampling rate of the cartesian coordinates in input grid pixels
        dy : float
            y-sampling rate of the cartesian coordinates in input grid pixels
        img_pr : numpy.ndarray(shape=(nr,np),dtype=float)
            output image
        r0 : float
            start radius in from the cartesian origin in scaled units
        p0 : float
            start azimuth from the cartesian x-axis in radians
        dr : float
            step size of the radial axis in the output in scaled units
        dp : float
            step size of the azimuth axis in the output in radians
        acc_norm : boolean, default: False
            flag to accumulate area normalization factors in the output
        
    """
    tpi = 2. * np.pi
    nyx = img_xy.shape
    #print("polar_rebin_core: ndim1:", nyx)
    nrp = img_pr.shape
    #print("polar_rebin_core: ndim2:", nrp)
    numr = nrp[0]
    nump = nrp[1]
    r1 = r0 + numr * dr
    p1 = p0 + nump * dp
    #print("polar_rebin_core: radial: ", [r0, r1, dr])
    #print("polar_rebin_core: azimuth: ", [p0, p1, dp])
    ix0 = min(nyx[1]-1, max(0, int(np.floor((x0 * dx - r1) / dx)))) # left pixel range clipped to image
    ix1 = min(nyx[1]-1, max(0, int(np.ceil((x0 * dx + r1) / dx)))) # right pixel range clipped to image
    iy0 = min(nyx[0]-1, max(0, int(np.floor((y0 * dy - r1) / dy)))) # bottom pixel range clipped to image
    iy1 = min(nyx[0]-1, max(0, int(np.ceil((y0 * dy + r1) / dy)))) # top pixel range clipped to image
    #print("polar_rebin_core: roi:", [[ix0,iy0],[ix1,iy1]])
    r = 0.
    phi = 0.
    img_pr[:,:] = 0.
    #afac0 = (dx * dy) / (dr * dp)
    apix = dx * dy # pixel area
    for j in range(iy0, iy1+1):
        y = (j - y0) * dy
        ysqr = y * y
        for i in range(ix0, ix1+1):
            x = (i - x0) * dx
            r = np.sqrt(x * x + ysqr)
            if (r < r0 or r > r1): continue # skip pixels out of radial range
            phi = np.arctan2(y, x)
            if phi < 0: phi += tpi # bring phi to positive range
            if (phi < p0 or phi > p1): continue # out of azimuth range
            jr = int((r - r0) / dr) # determine index of the radial bin, jr = 0 is from r = r0 to r0 + dr
            jp = int(np.round((phi - p0) / dp)) # determine index of the azimuthal bin, jp = 0 is from phi = 0 to phi = dp
            if (jr >= 0 and jr < numr and jp >= 0 and jp < nump):
                if acc_norm:
                    img_pr[jr,jp] += apix # accumulate the input area contributing to the polar bin
                else:
                    img_pr[jr,jp] += img_xy[j,i] # accumulate the flux density
#
@numba.jit
def polar_normalize_core(img_pr, r_0, d_r, d_p, afac):
    """

    Polar binning renormalization interface for the numba JIT compiler.
    
    Parameters
    ----------
        img_pr : numpy.ndarray(shape=(nr,np),dtype=float)
            in-/output image to be renormalized
        r0 : float
            start radius in from the cartesian origin in scaled units
        dr : float
            step size of the radial axis in the output in scaled units
        dp : float
            step size of the azimuth axis in the output in radians
        afac : numpy.ndarray(shape=(nr,np),dtype=float)
            normalization factors defining the area that contributed
        
    """
    nrp = img_pr.shape
    numr = nrp[0]
    nump = nrp[1]
    arp = d_r * d_p # common poler factor
    for jr in range(0, numr):
        rp = r_0 + (1.0 * jr + 0.5) * d_r # this is the bin mean radius
        for jp in range(0, nump):
            if afac[jr,jp] > 0.:
                img_pr[jr,jp] = img_pr[jr,jp] * arp * rp / afac[jr,jp]
#
def polar_rebin(image, num_rad, num_phi, pole, rng_rad, rng_phi = [0., 2.*np.pi], image_scale = [1., 1.], area_norm = False):
    """

    Transforms a 2D images to new grid in polar representation using a re-binning algorithm.

    Parameters
    ----------
        image : numpy.array floats with shape (...,ny,nx)
            data on a regular grid
        num_rad : integer
            number of radial sample
        num_phi : integer
            number of azimuthal samples
        pole : 2d tuples
            positions (x,y) of the pole in the input image (fractional pixel coordinate)
        rng_rad : float array of length 2
            radial range start and coverage (r_0, delta_r)
        rng_phi : float array of length 2
            azimuthal range start and coverage (phi_0, delta_phi) in radians
            default = [0., 2.*np.pi]
        image_scale : float array of length 2
            scale (x, y) of the input image in physical units per pixel, e.g. nm/pixel
            default = [1., 1.]
        area_norm : boolean, default: False
            flags area re-normalization. This interprets the input as differential
            f(x,y) dx dy in cartesian coordinates. Accumulation into each bin thus
            corresponds an integral over an area N dx dy, where N is the number of
            pixels added to a bin (it might be zero). The polar bin, however has
            an aera of r dr dphi, where dr is the radial step and dphi is the
            azimuthal step in radians. The output is then scaled by the fraction
            (r dr dphi) / (N dx dy). If N == 0, no scaling is applied or needed.

    Returns
    -------
        numpy.array of dimensions, shape (...,num_rad,num_phi)

    Remarks
    -------
        * Can be used to process on the last two dimensions of an array of
          higher dimensions.
        * Due to the non-isotropic distribution of polar bins, some output bins
          may not receive data, especially around the pole region. If you want
          to avoid this, you may use the routine polar_resample.
        * If the scale argument is used, the rng_rad input is assumed to be meant
          on the same scale as the input scale, i.e. the same physical scale unit.
          Also the azimuth argument is assumed to address physical angles, which may
          deviate from angles of a regular grid.
        * Calls a subroutine with @jit decorator. This does require longer excecution on
          the first call for run-time compiling, but will be faster afterwards.
        
    """
    nd = image.shape
    assert len(nd)>=2, 'need at least a 2d array'
    ld = len(nd)
    nx = nd[ld-1]
    ny = nd[ld-2]
    nimg = int(image.size / nx / ny)
    img = image.reshape(nimg, ny, nx) # reshape to a sequence of 2d images
    assert num_rad>=1, 'number of radial samples must be at least 1'
    assert num_phi>=1, 'number of azimuthal samples must be at least 1'
    npol = int(np.array(pole).size / 2)
    p_org = np.array(pole).reshape(npol,2) # sequence of pole positions
    assert (npol==nimg or npol==1), 'invalid number of pole positions'
    assert len(rng_rad)==2, 'invalid length of radial range input'
    assert np.abs(rng_rad[1])>0., 'invalid radial range parameter'
    assert len(rng_phi)==2, 'invalid length of azimuthal range input'
    assert np.abs(rng_phi[1])>0., 'invalid azimuth range parameter'
    assert len(image_scale)>=2, 'invalid length of input image scale parameter'
    assert (np.abs(image_scale[0])>0. and np.abs(image_scale[1])>0.), 'invalid input image scale'
    # initialize
    r_0 = rng_rad[0]
    d_r = rng_rad[1] / num_rad
    p_0 = rng_phi[0]
    d_p = rng_phi[1] / num_phi
    d_x = image_scale[0]
    d_y = image_scale[1]
    image_out = np.zeros((nimg,num_rad,num_phi),dtype=float) # polar data
    norm_pr = np.zeros((num_rad,num_phi),dtype=float)
    x_0 = p_org[0,0]
    y_0 = p_org[0,1]
    if area_norm: # get the area factors
        polar_rebin_core(img[0,:,:], x_0, y_0, d_x, d_y, norm_pr[:,:], r_0, p_0, d_r, d_p, acc_norm=True)
    for i in range(nimg):
        if (npol>1):
            x_0 = p_org[i,0]
            y_0 = p_org[i,1]
            if area_norm: # get the local area factors since the origin is volatile for a series
                polar_rebin_core(img[i,:,:], x_0, y_0, d_x, d_y, norm_pr[:,:], r_0, p_0, d_r, d_p, acc_norm=True)
        polar_rebin_core(img[i,:,:], x_0, y_0, d_x, d_y, image_out[i,:,:], r_0, p_0, d_r, d_p) # get flux density integral
        if area_norm: # area normalization
            polar_normalize_core(image_out[i,:,:], r_0, d_r, d_p, norm_pr[:,:])
    nout = np.array(nd).astype(int)
    nout[ld-2] = num_rad
    nout[ld-1] = num_phi
    return image_out.reshape(nout)
#
@numba.jit
def polar_rebin_rpoly3(img_xy, px0, py0, img_pr, r0, p0, dr, dp, x0, x1, x2, x3):
    """

    Transforms a 2D image to new grid in polar representation with the radial direction
    sampling on a grid with x = x0 + x1 * r + x2 * r**2 + x3 + r**3, where r is the distance
    to the pole in the input image and x the coordinate of the output image along axis = 0.
    The new x axis will correspond to a finer step size towards outer regions of the
    input image causing the respective data to be distributed more sparsely in the output.
    The transformation is based on a rebinning algorithm, which may lead to empty bins if
    the output grid has a higher density than the input grid.

    Parameters
    ----------
        img_xy : numpy.array of 2 dimensions, type float
            data on a regular grid
        px0, py0 : float
            position of the pole in the input image (fractional pixel coordinates)
        img_pr : numpy.array of 2 dimensions, type float
            outout polar grid
        r0 : float
            radial grid offset in input
        p0 : float
            azimuthal grid offset in input
        dr : float
            radial grid step size in input
        dp : float
            azimuthal grid step size in input
        x0 : float
            radial sampling offset, default = 0.
        x1 : float
            radial sampling linear coefficient, default = 1.
        x2 : float
            radial sampling 2nd order coefficient, default = 0.
        x3 : float
            radial sampling 3rd order coefficient, default = 0.
    
    Remarks
    -------
        * This routine has a @jit decorator. It does require longer excecution on
          the first call for run-time compiling, but will be faster afterwards.

    """
    nyx = img_xy.shape
    nrp = img_pr.shape
    tpi = 2.*np.pi # 2*Pi
    r1 = r0 + nrp[0] * dr
    p1 = p0 + nrp[1] * dp
    xl0 = x0 + x1 * r0 + x2 * r0**2 + x3 * r0**3 # lower radial limit
    xl1 = x0 + x1 * r1 + x2 * r1**2 + x3 * r1**3 # upper radial limit
    dxl = (xl1 - xl0) / nrp[0] # x sampling steps produced on the radial output dimension
    ix0 = min(nyx[1]-1, max(0, int(np.floor(px0 - r1)))) # left pixel range clipped to image
    ix1 = min(nyx[1]-1, max(0, int(np.ceil(px0 + r1)))) # right pixel range clipped to image
    iy0 = min(nyx[0]-1, max(0, int(np.floor(py0 - r1)))) # bottom pixel range clipped to image
    iy1 = min(nyx[0]-1, max(0, int(np.ceil(py0 + r1)))) # top pixel range clipped to image
    #
    # accumulate polar samples to polar bins
    r = 0.
    phi = 0.
    for j in range(iy0, iy1+1): # loop over input grid rows
        y = j - py0
        ysqr = y * y
        for i in range(ix0, ix1+1): # loop over input grid columns
            x = i - px0
            r = np.sqrt(x * x + ysqr)
            if (r < r0 or r > r1): continue # outside radial range, skip pixel
            phi = np.arctan2(y, x)
            if phi < 0: phi += tpi # wrap to range [0, 2 pi]
            if (phi < p0 or phi > p1): continue # out of azimuth range
            xl = x0 + x1 * r + x2 * r**2 + x3 * r**3 # corresponding output grid coordinate
            jr = int(np.round((xl - xl0) / dxl)) # round to next radial bin
            jp = int(np.round((phi - p0) / dp)) # round to next azimuthal bin
            if (jr >= 0 and jr < nrp[0] and jp >= 0 and jp < nrp[1]):
                img_pr[jr,jp] += img_xy[j,i]
    return
#
def radial_bin_mask(ndim, step, pix_pole, nr, radial_range=None):
    """

    Creates a mask linking image pixels to radial bins.

    Parameters
    ----------
        ndim : array of length 2 and type int
            dimensions of input images (num_rows, num_columns)
        step : array of length 2 and type float
            step sizes of input image (step_rows, step_columns)
        pix_pole : array of length 2 and type float
            pixel position of the pole from measuring radius (row, column)
        nr : int
            number of radial samples
        radial_range : array of lenth 2 and type float
            radial range, if None a default [0, nr] is used

    Returns
    -------
        array dimension ndim and type int containing indices of
        the radial 1d array linked to each input image

    """
    #
    msk = np.zeros(ndim, dtype=int)
    #
    if radial_range is None:
        r0 = 0.
        r1 = 1.0 * nr
    else:
        r0 = radial_range[0]
        r1 = radial_range[1]
    #
    for i_y in range(0, ndim[0]):
        y = step[0] * (i_y - pix_pole[0])
        for i_x in range(0, ndim[1]):
            x = step[1] * (i_x - pix_pole[1])
            r = np.sqrt(x**2 + y**2)
            i_r = round((r - r0) / (r1 - r0) * nr)
            if (i_r >= 0) and (i_r < nr):
                msk[i_y,i_x] = i_r
            else:
                msk[i_y,i_x] = -1
    #
    return msk



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 12 15:57:00 2019
@author: ju-bar

Peak functions and functions calculating their jacobians

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
import numpy as np
# %%
def gauss_2d(x_tuple, x0, y0, a, bxx, bxy, byy, c):
    '''
    a * Exp[ - (bxx*bxx + bxy*bxy) * (x-x0)**2 
             - 2 * bxy * (bxx + byy) * (x-x0) * (y-y0)
             - (byy*byy + bxy*bxy) * (y-y0)**2 ] + c
    '''
    (x, y) = x_tuple
    dx = x - x0
    dy = y - y0
    fxx = (bxx * bxx + bxy * bxy)
    fxy = 2 * bxy * (bxx + byy)
    fyy = (byy * byy + bxy * bxy)
    varg = fxx * dx**2 + 2 * fxy * dx * dy + fyy * dy**2
    return a * np.exp(-varg) + c
def gauss_2d_jac(x_tuple, x0, y0, a, bxx, bxy, byy, c):
    '''
    Derivatives of
        a * Exp[ - (bxx*bxx + bxy*bxy) * (x-x0)**2 
             - 2 * bxy * (bxx + byy) * (x-x0) * (y-y0)
             - (byy*byy + bxy*bxy) * (y-y0)**2 ] + c
    with respect to parameters x0, y0, a, bxx, bxy, byy, c
    Returns
        numpy.array([dx0, dy0, da, dbxx, dbxy, dbyy, dc]).T
    computed for all M input x_tuple sets, where x0 etc. can
    be arrays of length M
    '''
    (x, y) = x_tuple
    dx = x - x0
    dy = y - y0
    fxx = (bxx * bxx + bxy * bxy)
    fxy = 2 * bxy * (bxx + byy)
    fyy = (byy * byy + bxy * bxy)
    dx2 = dx**2
    dy2 = dy**2
    dxy = dx * dy
    varg = fxx * dx2 + 2 * fxy * dxy + fyy * dy2
    vexp = np.exp(-varg)
    dx0 = a * vexp * 2 * (fxx * dx + fxy * dy)
    dy0 = a * vexp * 2 * (fxy * dx + fyy * dy)
    da = vexp
    dbxx = -2 * a * vexp * (bxx * dx2 + bxy * dxy)
    dbxy = -2 * a * vexp * (bxy * (dx2 + dy2) + (bxx + byy) * dxy)
    dbyy = -2 * a * vexp * (bxy * dxy + byy * dy2)
    dc = np.full(np.size(x), 1., dtype=float)
    return np.array([dx0, dy0, da, dbxx, dbxy, dbyy, dc]).T
# %%
def gauss_2d_round(x_tuple, x0, y0, a, b, c):
    '''
    a * exp[ -b**2 * ((x-x0)**2 + (y-y0)**2) ] + c
    '''
    (x, y) = x_tuple
    dx = x - x0
    dy = y - y0
    varg = b**2 * (dx**2 + dy**2)
    return a * np.exp(-varg) + c
def gauss_2d_round_jac(x_tuple, x0, y0, a, b, c):
    '''
    Derivatives of 
        a * exp[ -b**2 * ((x-x0)**2 + (y-y0)**2) ] + c
    with respect to parameters x0, y0, a, b, c

    Returns
        numpy.array([dx0, dy0, da, db, dc]).T
    computed for all M input x_tuple sets, where x0 etc. can
    be arrays of length M
    '''
    (x, y) = x_tuple
    dx = x - x0
    dy = y - y0
    varg = b**2 * (dx**2 + dy**2)
    vexp = np.exp(-varg)
    dx0 = a * vexp * 2 * b**2 * dx
    dy0 = a * vexp * 2 * b**2 * dy
    da = vexp
    db = -2 * vexp * b * (dx**2 + dy**2)
    dc = np.full(np.size(x), 1., dtype=float)
    return np.array([dx0, dy0, da, db, dc]).T
# %%
def gauss_2d_round_slope(x_tuple, x0, y0, a, b, c, dx, dy):
    '''
    a * exp[-b**2 * ((x-x0)**2 + (y-y0)**2)] + c
        + dx*(x-x0) + dy*(y-y0)
    '''
    (x, y) = x_tuple
    gr = gauss_2d_round(x_tuple, x0, y0, a, b, c)
    return gr + dx*(x-x0) + dy*(y-y0)
def gauss_2d_round_slope_jac(x_tuple, x0, y0, a, b, c, dx, dy):
    '''
    Derivatives of 
        a * exp[-b**2 * ((x-x0)**2 + (y-y0)**2)] + c
        + dx*(x-x0) + dy*(y-y0)
    with respect to parameters x0, y0, a, b, c, dx, dy

    Returns
        numpy.array([dx0, dy0, da, db, dc, ddx, ddy]).T
    computed for all M input x_tuple sets, where x0 etc. can
    be arrays of length M
    '''
    (x, y) = x_tuple
    mx = x - x0
    my = y - y0
    varg = b**2 * (mx**2 + my**2)
    vexp = np.exp(-varg)
    dx0 = a * vexp * 2 * b**2 * mx
    dy0 = a * vexp * 2 * b**2 * my
    da = vexp
    db = -2 * vexp * b * (mx**2 + my**2)
    dc = np.full(np.size(x), 1., dtype=float)
    ddx = mx
    ddy = my
    return np.array([dx0, dy0, da, db, dc, ddx, ddy]).T
# %%



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 04 10:21:00 2021
@author: ju-bar

Modified on Mon Jun 20 11:48:00 2022 (ju-bar) added make_g_matrix, read_XTL

Functions handling input and output of structure data
via the XTL file format.

The XTL file format is a text format defining a volume in space
which is filled by atoms.

The supercell is defined by a sequence of numbers in the second
line of the file. The line starts with three numbers which set
the lengths a, b, and c of the cell edges in nanometers and
three more number setting the angles alpha, beta, and gamma
in degrees between these edges.

The third line is the electron beam energy in keV (for most)
of the programs reading XTL, but can be (in muSTEM) the number
of atom-type definitions following as sequency below. The
decision of which option is present in the file is made on reading
line four. Line four is a number in case line three is a voltage,
then line 4 is the number of atom types. Otherwise we find an
atom type symbol in line 4. Care must be taken when checking the
content of line four, since some strings, e.g. those containing a
letter "e" or "E" could be mis-interpreted as a float number.

An atom type definition consists of a variable number of text lines.
The first of these lines sets a name for the atom type, which is
can be used by the simulation program to distinguish types. This
is usually the element symbol, but can be used to distinguish
atomic sites occupied differently with the same element or for sites
of the same element but different thermal vibration. This string is
not used to determine the element type or atomic number, i.e. "R2D2"
would also be a valid input.
The second line defines properties of atomic sites for this atom type
using four numbers: (1) the number of atom sites (also the number of
lines following to define site positions), (2) the atomic number Z
(aka. core charge), (3) the occupancy factor (0.0 up to 1.0), and
(4) the mean squared displacement <u**2> in A**2 (<u**2> = B/(8 Pi**2),
where B is assumed to be an isotropic temperature factor, so that
<u**2> = (<u_x**2> + <u_y**2> + <u_z**2>)/3 with each Cartesian
component assumed to contribute equally. If ionic potentials are needed,
the charge of the ion can be added as fifth input, e.g. -1, +2, +1.3 etc.
What follows then are a number of lines equal to the first number of the
preceding line, which define atom positions by fractional coordinates
of the supercell, i.e. x/a  y/b  z/c.


Example:
# This is the first line of the example defining a BN crystal in [001] orientation
4.33704   2.50399   6.66120   90.0000   90.0000   90.0000
2
B
4    5.0000    1.000     0.0121
     0.000000  0.000000  0.250000
     0.500000  0.500000  0.250000
     0.333333  0.000000  0.750000
     0.833333  0.500000  0.750000
N
4    7.0000    1.000     0.0092
     0.000000  0.000000  0.750000
     0.500000  0.500000  0.750000
     0.333333  0.000000  0.250000
     0.833333  0.500000  0.250000


This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

import re
import copy
import numpy as np
from emilys.structure.supercell import supercell
import emilys.structure.atom as ato

def make_g_matrix(g1, g2, imax):
    a1 = np.array(g1, dtype=int) # force basis vector of integer index
    a2 = np.array(g2, dtype=int) # force basis vector of integer index
    l_g = []
    l_l = []
    for j2 in range(-imax, imax+1):
        v2 = a2 * j2
        for j1 in range(-imax, imax+1):
            v1 = a1 * j1
            v = v1 + v2
            lv = np.sqrt(np.dot(v, v))
            li = -1
            if len(l_l) > 0:
                for i in range(0, len(l_l)):
                    if l_l[i] > lv:
                        li = i
                        break
            if li >= 0:
                l_g.insert(li, v)
                l_l.insert(li, lv)
            else:
                l_g.append(v)
                l_l.append(lv)
    return l_g

def write_XTL(sc, file, l_type_name_adds = [], d_adds = {}):
    """

    Writes atomic structure data in XTL file format.

    Parameters
    ----------

        sc : emilys.structure.supercell.supercell
            Supercell data the should be written to file.
        file : str
            File name
        l_type_name_adds : list
            List of strings that determine additions made to the
            type names:
            'occ' : adds occupancy
            'uiso' : adds the thermal vibration mean square amplitude
            'ion' : adds the ionic charge, also triggers ionic charge output
        d_adds : dictionary
            Dictionary for adding additional information to the XTL file.
            Supported keys are
            'ht' : adds a high tension value (kV) in the third line
            'bragg' : adds scan and bragg vector lists, sub-list
                'sz' : zone axis vector, 3d-vector, e.g. [0, 0, 1]
                'sx' : scan vector 1, 3d-vector, e.g. [1, 0, 0]
                'sy' : scan vector 2, 3d-vector, e.g. [0, 1, 0]
                'g1' : bragg basis 1, 3d-vector, e.g. [1, 1, 0]
                    perp to sz
                'g2' : bragg basis 2, 3d-vector, e.g. [1, -1, 0]
                    perp to sz, not colinear to g1
                'gi_max' : max. bragg order to include, e.g. 5,
                    this spans a matrix by g1 and g2 up to the given number of multiples
                'gi_lim' : limits the length of the bragg vector list
    
    Returns
    -------

        int
            Error code

    """
    io_err = 0
    ionic = False
    assert isinstance(sc, supercell), 'This expects that sc is input of type emilys.structure.supercell.supercell'
    assert isinstance(file, str), 'This expects that file is input of type str'
    if 'ion' in l_type_name_adds:
        ionic = True
    with open(file, "w") as file_out:
        file_out.write(sc.get_composition_str() + " (xtlio)\n")
        file_out.write("{:<11.6f}{:<11.6f}{:<11.6f}{:<10.4f}{:<10.4f}{:<10.4f}\n".format(
            sc.a0[0], sc.a0[1], sc.a0[2], sc.angles[0], sc.angles[1], sc.angles[2]))
        if 'ht' in d_adds.keys(): # add high tension line
            file_out.write("{:.5f}\n".format(d_adds['ht']))
        d = sc.get_type_dict(None, l_type_name_adds)
        natty = len(d.keys())
        file_out.write("{:d}\n".format(natty)) # number of atom types
        for satty in d.keys():
            atty = d[satty]
            file_out.write(satty + "\n") # atom type name
            # atom type paramaters
            if ionic:
                sion = ato.get_str_from_charge(atty["ion"])
                file_out.write("{:<6d} {:<10.3f} {:<10.3f} {:<11.5E}  {:s}\n".format(len(atty["sites"]),atty["Z"],atty["occ"],atty["uiso"],sion)) # number of atom types
            else:
                file_out.write("{:<6d} {:<10.3f} {:<10.3f} {:<11.5E}\n".format(len(atty["sites"]),atty["Z"],atty["occ"],atty["uiso"])) # number of atom types
            # atom type sites
            for pos in atty["sites"]:
                file_out.write("       {:<10.6f} {:<10.6f} {:<10.6f}\n".format(pos[0], pos[1], pos[2]))
        if 'bragg' in d_adds.keys(): # add scan and bragg list
            d_b = d_adds['bragg']
            vsz = d_b['sz']
            vsx = d_b['sx']
            vsy = d_b['sy']
            gi_lim = 0
            if (('g1' in d_b.keys()) and ('g2' in d_b.keys()) and ('gi_max' in d_b.keys())):
                vg1 = d_b['g1']
                vg2 = d_b['g2']
                gi_max = d_b['gi_max']
                gi_lim = 0
                if 'gi_lim' in d_b.keys():
                    gi_lim = d_b['gi_lim']
                g_list = make_g_matrix(vg1, vg2, gi_max)
            file_out.write("{:d}\n".format(min(gi_lim, len(g_list))))
            file_out.write("   {:d}   {:d}   {:d}\n".format(vsz[0],vsz[1],vsz[2]))
            file_out.write("   {:d}   {:d}   {:d}\n".format(vsx[0],vsx[1],vsx[2]))
            file_out.write("   {:d}   {:d}   {:d}\n".format(vsy[0],vsy[1],vsy[2]))
            for i in range(0, min(gi_lim, len(g_list))):
                vg = g_list[i]
                file_out.write("   {:d}   {:d}   {:d}\n".format(vg[0],vg[1],vg[2]))
        file_out.write("\n")
        file_out.close()
    return io_err

def read_XTL(file, debug=False):
    """
    
    Reads atomic structure data from a file assuming the XTL file format.

    Parameters
    ----------

        file : str
            Input file name.
        debug : boolean, default: False
            Flags debug print out of the file parsing.

    Returns
    -------

        emilys.structure.supercell.supercell

    """
    assert type(file) is str, "This expects a string as input parameter."
    # open the file and read the lines
    with open(file) as file_in:
        if debug: print('dbg (read_xtl): opened file [' + file + ']')
        lines = []
        for line in file_in:
            lines.append(line)
        file_in.close()
    if debug: print('dbg (read_xtl): read ', len(lines),' lines of text.')
    assert len(lines) > 3, "The input file doesn't contain sufficient number of text lines."
    # create a supercell object
    sc = supercell()
    # process the cell parameters
    l_cmp = re.split(' +|,|;|\t+', lines[1].strip()) # decompose to list of string
    assert len(l_cmp) > 5, "The second line of the input file could not be split into >6 items."
    if debug: print('dbg (read_xtl): supercell input line: ', l_cmp)
    sc.a0 = np.array([float(l_cmp[0]),float(l_cmp[1]),float(l_cmp[2])]) # cell size in Angst
    if debug: print('dbg (read_xtl): supercell size [A]: a = {:.5f}, a = {:.5f}, c = {:.5f}'.format(*sc.a0))
    sc.angles = np.array([float(l_cmp[3]),float(l_cmp[4]),float(l_cmp[5])])
    if debug: print('dbg (read_xtl): supercell angles [deg]: alpha = {:.4f}, beta = {:.4f}, gamma = {:.4f}'.format(*sc.angles))
    natty = 0 # preset number of atom types
    # determine option of line 3 using also line 4
    if lines[3].strip().isdigit(): # line 4 is a number
        ht = float(lines[2].strip()) # -> line 3 is the beam energy (or high tension)
        sc.d_add['ht'] = ht # write to additional data dictionary
        if debug: print('dbg (read_xtl): added beam energy [keV]: {:.4f}'.format(ht))
        natty = int(lines[3].strip()) # get number of atom type definitions following line 4
        i_line = 4 # first line of atom type data
    else: # line 4 is not a number -> line 3 is the number of atom types
        natty = int(lines[2].strip()) # get number of atom type definitions following line 3
        i_line = 3 # first line of atom type data
    if debug: print('dbg (read_xtl): number of atom types:', natty)
    
    if natty > 0 and i_line < len(lines): # there are atom types to read
        for iatty in range(0, natty):
            s_aty_label = lines[i_line].strip() # read atom type label (not used further)
            i_line += 1
            assert i_line < len(lines), "More data expected (atom type definition), file seems broken. (line #{:d})".format(i_line)
            s_aty_def = lines[i_line].strip() # read the type definition string
            i_line += 1
            l_cmp = re.split(' +|,|;|\t+',s_aty_def) # decompose to list of strings
            nat = int(l_cmp[0])
            crg = 0.0
            if len(l_cmp) > 4:
                crg = float(l_cmp[4])
            a = ato.atom(Z=int(float(l_cmp[1])), occ=float(l_cmp[2]), uiso=float(l_cmp[3]), charge=crg)
            if nat > 0: # read atom positions
                for iat in range(0, nat):
                    assert i_line < len(lines), "More data expected (atom position), file seems broken. (line #{:d})".format(i_line)
                    l_pos = re.split(' +|,|;|\t+', lines[i_line].strip())
                    i_line += 1
                    assert len(l_pos)>=3, "Three position values (fx, fy, fz), file seems broken. (line #{:d})".format(i_line)
                    f_p = [float(l_pos[0]), float(l_pos[1]), float(l_pos[2])]
                    b = ato.atom(Z=a.Z, occ=a.occ, uiso=a.uiso, charge=a.charge, pos=f_p)
                    sc.l_atoms.append(b)
    if debug: print('dbg (read_xtl): added ', len(sc.l_atoms), ' atoms to the supercell.')
    return sc


# -*- coding: utf-8 -*-
"""
Created on Wed Sep 14 12:18:00 2022
@author: ju-bar

TTX file I/O

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""

def write_table_txt(str_filename, tab, header, str_sep = '  ', str_fmt = '{:12.6E}'):
    """

    Writes a text file which contains the data of tab as
    text table.

    Parameter
    ---------
        str_filename : str
            output file name
        tab : numpy.ndarray(shape=(num_cols,num_rows))
            input data as numpy array
        header : list
            list of strings defining the header to write
        str_sep : str, default = '  '
            value separator string
        str_fmt : str, default = '{:12.6E}'
            value format command for str.format()

    """
    ndim = tab.shape
    assert len(ndim) == 2, 'Error: expecting a 2-dimensional input'
    num_cols = ndim[0]
    num_rows = ndim[1]
    file = open(str_filename, 'w+')
    tab_t = tab.T # transposed table
    if header is not None: # write headers
        if len(header) > 0: # only one header line allowed
            s_line = ''
            for sh in header:
                if len(s_line) > 0: s_line += str_sep # add separator before next header item
                s_line += sh # add next header item
            file.write(s_line + '\n')
    if num_cols > 0:
        for irow in range(0, num_rows):
            s_line = ''
            for icol in range(0, num_cols):
                if len(s_line) > 0: s_line += str_sep # add separator before next header item
                s_line += str_fmt.format(tab_t[irow,icol]) # add next header item
            file.write(s_line + '\n')
    file.close()
    
